// ==============================================================
//
// as350.cpp
// Control module for as350 vessel class
//
// ==============================================================

#define STRICT 1
#define ORBITER_MODULE

#include "orbitersdk.h"
#include "OrbiterSoundSDK40.h"
#include <cstdio>
#include <cmath>
#include <string>
#include <stdio.h>

// ==============================================================
// Some vessel parameters
// ==============================================================
const DWORD COMM_SET_TGT = 2306;

#define STRT	1
#define MOTR	3

const double EMPTY_MASS = 1300; // inc pilot ;)
const double FUEL_MASS = 800; // inc auxiliary
const double MAIN_TH = 28000; // guess
const double MAIN_ISP = 400000; // guess
const double YAW_TH = 2500; // guess
const double RTR_OFF = 0.5;

OBJHANDLE hGref;

//// beacons
VECTOR3 bs0pos = _V(  0, 1.639, -7.319);
VECTOR3 bs0col = _V(1,0,0);
VECTOR3 bs1pos = _V( 1.22, 0.119, -4.936);
VECTOR3 bs1col = _V(0,1,0);
VECTOR3 bs2pos = _V(-1.22, 0.119, -4.936);
VECTOR3 bs2col = _V(0,1,0);
VECTOR3 bs3pos = _V( 0, -1.008, 2.182);
VECTOR3 bs3col = _V(1,1,1);

static const DWORD ntdvtx_skis = 7;
static TOUCHDOWNVTX tdvtx_skis[ntdvtx_skis] = {
	{_V(0, -1.45, 1.6), 1e6, 1e5, 2, 0.5},
	{_V(-1.25,-1.45,-0.9), 1e6, 1e5, 2, 0.5},
	{_V( 1.25,-1.45,-0.9), 1e6, 1e5, 2, 0.5},
	{_V(   0 , 1.3,  -7.5), 1e6, 1e5, 1},
	{_V(-1.25,-1.45, 1.6), 1e6, 1e5, 1},
	{_V( 1.25,-1.45, 1.6), 1e6, 1e5, 1},
	{_V( 0 , 1.3,     0), 1e6, 1e5, 1}
};


void VLiftCoeff (double aoa, double M, double Re, double *cl, double *cm, double *cd)
{
	static const double step = RAD*22.5;
	static const double istep = 1.0/step;
	static const int nabsc = 17;
	static const double CL[nabsc] = {0, 0.1, 0.2, 0.1, 0, 0.1, 0.2, 0.1, 0, -0.1, -0.2, -0.1, 0, -0.1, -0.2, -0.1, 0};


	aoa += PI;
	int idx = max (0, min (15, (int)(aoa*istep)));
	double d = aoa*istep - idx;
	*cl = 2*(CL[idx] + (CL[idx+1]-CL[idx])*d);
	*cm = 0.0;
	*cd = 5 + oapiGetInducedDrag (*cl, 0.16, 0.2);
	*cl = 0;
	
}

// 2. horizontal lift component (vertical stabiliser and body)

void HLiftCoeff (double beta, double M, double Re, double *cl, double *cm, double *cd)
{
	static const double step = RAD*22.5;
	static const double istep = 1.0/step;
	static const int nabsc = 17;
	static const double CL[nabsc] = {0, 0.1, 0.2, 0.1, 0.0, 0.1, 0.2, 0.1, 0, -0.1, -0.2, -0.1, 0,-0.1,-0.2,-0.1, 0};
	beta += PI;
	int idx = max (0, min (15, (int)(beta*istep)));
	double d = beta*istep - idx;
	*cl = -1*(CL[idx] + (CL[idx+1]-CL[idx])*d);
	*cm = 0.0;
	*cd = 1.1*(0.11 + oapiGetInducedDrag (*cl, 0.16, 0.2));
	//sprintf(oapiDebugString(), "beta %0.3f", beta*DEG);

}



// ==============================================================
// Shuttle-PB class interface
// ==============================================================

class as350: public VESSEL3 {
public:
	as350 (OBJHANDLE hVessel, int flightmodel);
	~as350 ();
	void clbkSetClassCaps (FILEHANDLE cfg);
	void clbkPreStep(double SimT, double SimDT, double MJD);
	//////void clbkPostStep (double simt, double simdt, double mjd);
	void clbkSaveState (FILEHANDLE scn);
	void clbkLoadStateEx (FILEHANDLE scn, void *status);
	int clbkConsumeBufferedKey(DWORD key, bool down, char *kstate);
	int clbkConsumeDirectKey(char *kstate);
	bool clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp);
	void clbkPostCreation();



private:

	void SetMainTh(void);
	VECTOR3 RotMat(VECTOR3 CURR_VECT,VECTOR3 ROT_AXIS,double rot_angle);
	void CheckForGrapple(void);
	void SetHookPos(void);
	void DoTgtInfo(void);
	void DoHoldPos(void);
	void DoHoldAlt(void);
	void DoBoom(void);
	void DefineAnimations (void);
	int MyID;
	char tgt_name[256];

	THRUSTER_HANDLE th_main[1], th_yaw[2], th_boom[2];
	THGROUP_HANDLE htgrpmain, th_group[1];
	PROPELLANT_HANDLE hProp;
	ATTACHMENTHANDLE hatp; 
	BEACONLIGHTSPEC beac0, beac1, beac2, beac3;

	ANIMATIONCOMPONENT_HANDLE Section0;
	MGROUP_TRANSFORM *rtr0, *rtr1, *rtr2, *rtr3, *rtr4, *rtr5, *rtr6, *rtr7, *rtr8, *rtr9, *hook1, *hook2;
	MGROUP_TRANSFORM *panel1, *panel2;

	int eng_status, hook_status, light_status, grapple_status, auto_hvr_status, auto_alt_status;
	double step_time, pitch_level, bank_level, yaw_level, rotor_trim, elev_trim;
	double drot, dthrottle, boom_sw;
	int pitch_mode, bank_mode, yaw_mode, throttle_mode;

	UINT anim_rtr1_rotb, anim_rtr1_rott, anim_rtr1_blds_scale, anim_rtr1_trns_scale;
	UINT anim_rtr2_rotb, anim_rtr2_rott, anim_rtr2_blds_scale, anim_rtr2_trns_scale;
	UINT anim_rtr_axle, anim_hook;
	UINT anim_rtr_tilt;
	UINT anim_panel1, anim_panel2;
	double rtr1_rot_proc, hook_proc;

	double ground_hdg, ground_spd, tgt_brg, tgt_dist, tgt_delta_alt;

	double js_pu, js_pd, js_bl, js_br, js_yl, js_yr;
	double js_pitch_level, js_bank_level, js_yaw_level;

};

as350::as350 (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
    

}

as350::~as350 ()
{
	delete rtr1;
	delete rtr2;
	delete rtr3;
	delete rtr4;
	delete rtr5;
	delete rtr6;
	delete rtr7;
	delete rtr8;
	delete rtr9;
	delete hook1;
	delete hook2;
	delete panel1;
	delete panel2;
}

void as350::DefineAnimations (void)
{
	//tilt rtr
	anim_rtr_tilt = CreateAnimation (0);
	static UINT RtrGrp10[2] = {7,19};
	rtr0 = new MGROUP_ROTATE (0, RtrGrp10, 0, _V(0,1.66,0), _V(1,0,0), (float)(10*RAD));
	Section0 = AddAnimationComponent (anim_rtr_tilt, 0, 1, rtr0);

// 1. RTR1 ROTb
	anim_rtr1_rotb = CreateAnimation (0);
	static UINT RtrGrp0[1] = {7};
	rtr1 = new MGROUP_ROTATE (0, RtrGrp0, 1, _V(0,0,0), _V(0,1,0), (float)(360*RAD));
	AddAnimationComponent (anim_rtr1_rotb, 0, 1, rtr1, Section0);

	//1b RTR1 ROTt
	anim_rtr1_rott = CreateAnimation (0);
	static UINT RtrGrp1[1] = {19};
	rtr2 = new MGROUP_ROTATE (0, RtrGrp1, 1, _V(0,0,0), _V(0,1,0), (float)(360*RAD));
	AddAnimationComponent (anim_rtr1_rott, 0, 1, rtr2, Section0);

	//1c RTR1 b scale
	anim_rtr1_blds_scale = CreateAnimation (0);
	static UINT RtrGrp2[1] = {7};
	rtr3 = new MGROUP_SCALE (0, RtrGrp2, 1, _V(0,1.66,0), _V(0.0001,0.0001,0.0001));
	AddAnimationComponent (anim_rtr1_blds_scale, 0, 1, rtr3);

	//1d RTR1 t scale
	anim_rtr1_trns_scale = CreateAnimation (0);
	static UINT RtrGrp3[1] = {19};
	rtr4 = new MGROUP_SCALE (0, RtrGrp3, 1, _V(0,1.66,0), _V(0.0001,0.0001,0.0001));
	AddAnimationComponent (anim_rtr1_trns_scale, 0, 1, rtr4);

	// axle
	anim_rtr_axle = CreateAnimation (0);
	static UINT RtrGrp4[1] = {2};
	rtr5 = new MGROUP_ROTATE (0, RtrGrp4, 1, _V(0,0,0), _V(0,1,0), (float)(360*RAD));
	AddAnimationComponent (anim_rtr_axle, 0, 1, rtr5);

// 2. RTR2 ROT
	anim_rtr2_rotb = CreateAnimation (0);
	static UINT RtrGrp5[1] = {8};
	rtr6 = new MGROUP_ROTATE (0, RtrGrp5, 1, _V(0,0.306,-6.14), _V(1,0,0), (float)(360*RAD));
	AddAnimationComponent (anim_rtr2_rotb, 0, 1, rtr6);

	//2b RTR2 ROTt
	anim_rtr2_rott = CreateAnimation (0);
	static UINT RtrGrp6[1] = {18};
	rtr7 = new MGROUP_ROTATE (0, RtrGrp6, 1, _V(0,0.306,-6.14), _V(1,0,0), (float)(360*RAD));
	AddAnimationComponent (anim_rtr2_rott, 0, 1, rtr7);

	//2c RTR2 b scale
	anim_rtr2_blds_scale = CreateAnimation (0);
	static UINT RtrGrp7[1] = {8};
	rtr8 = new MGROUP_SCALE (0, RtrGrp7, 1, _V(0,0.306,-6.14), _V(0.0001,0.0001,0.0001));
	AddAnimationComponent (anim_rtr2_blds_scale, 0, 1, rtr8);

	//1d RTR1 t scale
	anim_rtr2_trns_scale = CreateAnimation (0);
	static UINT RtrGrp8[1] = {18};
	rtr9 = new MGROUP_SCALE (0, RtrGrp8, 1, _V(0,0.306,-6.14), _V(0.0001,0.0001,0.0001));
	AddAnimationComponent (anim_rtr2_trns_scale, 0, 1, rtr9);

	///////// hook
	anim_hook = CreateAnimation (0);
	static UINT HkGrp0[1] = {9};
	hook1 = new MGROUP_ROTATE (0, HkGrp0, 1, _V(0, -1.193, 0.737), _V(1,0,0), (float)(105*RAD));
	Section0 = AddAnimationComponent (anim_hook, 0, 1, hook1);

	static UINT HkGrp1[1] = {10};
	hook2 = new MGROUP_TRANSLATE (0, HkGrp1, 1, _V(0, 0, 1.5));
	AddAnimationComponent (anim_hook, 0, 1, hook2, Section0);

	////// panels night/day
	anim_panel1 = CreateAnimation (0);
	static UINT PanelGrp1[2] = {12,16};
	panel1 = new MGROUP_SCALE (0, PanelGrp1, 2, _V(0,-0.571,3.074), _V(0.0001,0.0001,0.0001));
	AddAnimationComponent (anim_panel1, 0, 1, panel1);

	anim_panel2 = CreateAnimation (0);
	static UINT PanelGrp2[2] = {6,15};
	panel2 = new MGROUP_SCALE (0, PanelGrp2, 2, _V(0,-0.571,3.074), _V(0.0001,0.0001,0.0001));
	AddAnimationComponent (anim_panel2, 0, 1, panel2);


}


// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void as350::clbkSetClassCaps (FILEHANDLE cfg)
{
	strcpy(tgt_name, "NULL");
	eng_status = 0;
	step_time = 0;
	pitch_level = 0;
	bank_level = 0;
	yaw_level = 0;
	rtr1_rot_proc = 0;
	hook_proc = 0;
	hook_status = 0;
	pitch_mode = 0;
	bank_mode = 0;
	yaw_mode = 0;
	throttle_mode = 0;
	dthrottle = 0;

	js_pu = 0; // joystick input
	js_pd = 0;
	js_bl = 0;
	js_br = 0;
	js_yl = 0;
	js_yr = 0;
	js_pitch_level = 0;
	js_bank_level = 0;
	js_yaw_level = 0;

	rotor_trim = 0;
	elev_trim = 0;
	drot = 0;
	light_status = 0;
	grapple_status = 0;

	auto_hvr_status = 0;
	auto_alt_status = 0;

	ground_hdg = 0;
	ground_spd = 0;

	tgt_brg = 0;
	tgt_dist = 0;
	tgt_delta_alt = 0;
	boom_sw = 0;

	DefineAnimations ();
	// physical vessel parameters
	SetSize (4);
	SetEmptyMass (EMPTY_MASS);
	SetPMI(_V(4,6,4.5));
	SetCrossSections (_V(10,10,3));
	SetCW(1,1,1,1);
	SetRotDrag (_V(0.5,0.5,0.5));
	SetTouchdownPoints ( tdvtx_skis, ntdvtx_skis);

	// camera parameters
	SetCameraOffset (_V( 0.25, 0.225, 1.625));
	SetMeshVisibilityMode(AddMesh("GenesisMission/as350"),MESHVIS_ALWAYS);
	
	

	// ***************** thruster definitions *******************

	hProp = CreatePropellantResource (FUEL_MASS,FUEL_MASS);
	
	th_main[0] = CreateThruster (_V( 0, RTR_OFF, 0), _V( 0,1,0), MAIN_TH, hProp, MAIN_ISP);

	th_group[0] = th_main[0];
	htgrpmain = CreateThrusterGroup (th_group, 1, THGROUP_MAIN);

	th_yaw[0] = CreateThruster (_V( 0, 0, -3), _V( 1,0,0), YAW_TH, hProp, MAIN_ISP);
	th_yaw[1] = CreateThruster (_V( 0, 0, -3), _V(-1,0,0), YAW_TH, hProp, MAIN_ISP);

	th_group[0] = th_yaw[0];
	CreateThrusterGroup (th_group, 1, THGROUP_ATT_YAWLEFT);

	th_group[0] = th_yaw[1];
	CreateThrusterGroup (th_group, 1, THGROUP_ATT_YAWRIGHT);

	///////////////////////////////////////////////////////
	/////AirFoils

	CreateAirfoil2 (LIFT_VERTICAL, _V(0,-0.0,0), VLiftCoeff, 1, 1, 0.5);
	CreateAirfoil2 (LIFT_HORIZONTAL, _V(0,0,-4), HLiftCoeff, 1, 2, 1);

	CreateControlSurface3(AIRCTRL_ELEVATORTRIM, 0.5, 2, _V(0,0,-3), AIRCTRL_AXIS_XPOS);

	/////////////// attach pnt
	hatp = CreateAttachment (false, _V(0,-1.193, 4.45), _V(0,0,1), _V(0,1,0), "AS");

	////// beacons
	beac0.shape = BEACONSHAPE_STAR;
	beac0.pos = &bs0pos;
	beac0.col = &bs0col;
	beac0.size = 0.3;
	beac0.falloff = 0.75;
	beac0.period = 2.0;
	beac0.duration = 0.05;
	beac0.tofs = 1.0;
	beac0.active = false;
	AddBeacon (&beac0);

	beac1.shape = BEACONSHAPE_STAR;
	beac1.pos = &bs1pos;
	beac1.col = &bs1col;
	beac1.size = 0.3;
	beac1.falloff = 0.75;
	beac1.period = 2.0;
	beac1.duration = 0.05;
	beac1.tofs = 0.0;
	beac1.active = false;
	AddBeacon (&beac1);

	beac2.shape = BEACONSHAPE_STAR;
	beac2.pos = &bs2pos;
	beac2.col = &bs2col;
	beac2.size = 0.3;
	beac2.falloff = 0.75;
	beac2.period = 2.0;
	beac2.duration = 0.05;
	beac2.tofs = 0.0;
	beac2.active = false;
	AddBeacon (&beac2);

	beac3.shape = BEACONSHAPE_STAR;
	beac3.pos = &bs3pos;
	beac3.col = &bs3col;
	beac3.size = 0.3;
	beac3.falloff = 0.75;
	beac3.period = 2.0;
	beac3.duration = 0.05;
	beac3.tofs = 1.0;
	beac3.active = false;
	AddBeacon (&beac3);

	///////////////////////////// boom p streams

	th_boom[0] = CreateThruster (_V( 0, 0, 0), _V( 0,0,1), 0.001, hProp, 2, 1);
	th_boom[1] = CreateThruster (_V( 0, 0, 0), _V( 0,0,1), 0.001, hProp, 2, 1);

	SURFHANDLE boom1 = oapiRegisterParticleTexture ("Contrail4");
	SURFHANDLE boom2 = oapiRegisterParticleTexture ("Contrail3");

	PARTICLESTREAMSPEC boom_pst1 = {
		0, 8, 100, 150, 2, 2, 10, 4, PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-7, 1,
		boom1
	};

	PARTICLESTREAMSPEC boom_pst2 = {
		0, 8, 80, 100, 2, 1, 15, 2, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-7, 1,
		boom2
	};

	AddExhaustStream(th_boom[0],_V(0,0,0),&boom_pst1);
	AddExhaustStream(th_boom[1],_V(0,0,0),&boom_pst2);
}



// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new as350 (hvessel, flightmodel);
}

// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (as350*)vessel;
}




void as350::clbkSaveState (FILEHANDLE scn)
{
	char cbuf[256];
	elev_trim = GetControlSurfaceLevel(AIRCTRL_ELEVATORTRIM);

	// default vessel parameters
	VESSEL3::clbkSaveState(scn);

	sprintf (cbuf, "%0.4f %0.4f %0.4f", rtr1_rot_proc, hook_proc, drot);
	oapiWriteScenario_string (scn, "ANMT", cbuf);

	sprintf (cbuf, "%d %d %d %0.4f %0.4f", eng_status, hook_status, light_status, rotor_trim, elev_trim);
	oapiWriteScenario_string (scn, "PRMS", cbuf);

	sprintf (cbuf, "%s %d %d %0.4f", tgt_name, auto_hvr_status, auto_alt_status, boom_sw);
	oapiWriteScenario_string (scn, "NAVS", cbuf);

}


void as350::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
    char *line;
	while (oapiReadScenario_nextline (scn, line)) {
		if (!strnicmp (line, "ANMT", 4)) 
		{
			sscanf (line+4, "%lf%lf%lf", &rtr1_rot_proc, &hook_proc, &drot);
		} 
		else if (!strnicmp (line, "PRMS", 4)) 
		{
			sscanf (line+4, "%d%d%d%lf%lf", &eng_status, &hook_status, &light_status, &rotor_trim, &elev_trim);
		}
		else if (!strnicmp (line, "NAVS", 4)) 
		{
			sscanf (line+4, "%s%d%d%lf", &tgt_name, &auto_hvr_status, &auto_alt_status, &boom_sw);
		} 
		else 
		{
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
   }
	
	SetControlSurfaceLevel(AIRCTRL_ELEVATORTRIM, elev_trim, true);
	SetAnimation(anim_hook, hook_proc);
	SetHookPos();
}

bool SetTgt(void *id, char *str, void *data)
{
	char cbuf[256];
	sscanf(str, "%s", &cbuf);
	if (!strnicmp (cbuf, "NULL", 4)) // accept no target
		{
		VESSEL2* vme = (VESSEL2*)oapiGetVesselInterface(oapiGetFocusObject ());
		vme->SendBufferedKey(COMM_SET_TGT, false, cbuf);
		return true;
		}
	else 
		{
		OBJHANDLE bbtgt = oapiGetObjectByName(cbuf);
		if (bbtgt == NULL)bbtgt = oapiGetBaseByName(hGref, cbuf);
		if (bbtgt == NULL)return false; // no such object
		else 
			{
			VESSEL2* vme = (VESSEL2*)oapiGetVesselInterface(oapiGetFocusObject ());
			vme->SendBufferedKey(COMM_SET_TGT, false, cbuf);
			return true;
			}
		}
}

int as350::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate)
{
	if (key == COMM_SET_TGT) 
	{
		sscanf (kstate, "%s", &tgt_name);
		return 1;
	}

  if (!down) return 0; // only process keydown events

 if (KEYMOD_SHIFT (kstate)) 
  {	 
	switch (key)
	{
		

	} 
  }


 else if (KEYMOD_CONTROL (kstate)) 
  { 
	switch (key)
	{
		
	} 
  }

else { // unmodified keys
    switch (key) 
		{
		case OAPI_KEY_K: // engine start / stop
		if(boom_sw == 0)
		{
		if (eng_status == 0) eng_status = 1; // wind up
		else if (eng_status == 1) eng_status = 3; // wind down
		else if (eng_status == 2) eng_status = 3;
		else if (eng_status == 3) eng_status = 1;
		}
		return 1;

		case OAPI_KEY_G: // hook
		if (hook_status == 0) hook_status = 1;
		else if (hook_status == 1) hook_status = 3;
		else if (hook_status == 2) hook_status = 3;
		else if (hook_status == 3) hook_status = 1;
		return 1;

		case OAPI_KEY_V: // auto hover
		if (auto_hvr_status == 0) 
		{	
			auto_hvr_status = 1;
			auto_alt_status = 0;
		}
		else if (auto_hvr_status == 1)
		{
			SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT,0);
			SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT,0);
			auto_hvr_status = 0;
		}
		return 1;

		case OAPI_KEY_B: // auto hold alt
		if (auto_alt_status == 0) 
		{
			auto_alt_status = 1;
			auto_hvr_status = 0;
			SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT,0);
			SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT,0);
		}
		else if (auto_alt_status == 1)auto_alt_status = 0;
		
		return 1;

		case OAPI_KEY_J: // hook
		if (grapple_status == 2)
			{
			if(DetachChild(hatp) == true)grapple_status = 3;
			}
		return 1;

		case OAPI_KEY_P:
		hGref = GetGravityRef();
		oapiOpenInputBox("Enter Target Name",SetTgt, 0, 50, 0);
		return 1;

		case OAPI_KEY_L: // lights
		RESETKEY (kstate, OAPI_KEY_L);
		if (light_status == 0) 
			{
			light_status = 1;
			SetAnimation(anim_panel1, 1);
			SetAnimation(anim_panel2, 0);
			beac0.active = true;
			beac1.active = true;
			beac2.active = true;
			beac3.active = true;
			}
		else if (light_status == 1)
			{
			light_status = 0;
			SetAnimation(anim_panel1, 0);
			SetAnimation(anim_panel2, 1);
			beac0.active = false;
			beac1.active = false;
			beac2.active = false;
			beac3.active = false;
			}
		return 1;
			
			case OAPI_KEY_NUMPAD5: 
			RESETKEY (kstate, OAPI_KEY_NUMPAD5);
			return 1;
		}
  }
  return 0;
}


int as350::clbkConsumeDirectKey(char *kstate)
{
	if (KEYMOD_SHIFT (kstate))
	{
		if ((auto_alt_status + auto_hvr_status) == 0)
		{
		if (KEYDOWN (kstate, OAPI_KEY_ADD)) 
			{
			RESETKEY (kstate, OAPI_KEY_ADD);
			throttle_mode = 1;
			}

		if (KEYDOWN (kstate, OAPI_KEY_SUBTRACT)) 
			{
			RESETKEY (kstate, OAPI_KEY_SUBTRACT);
			throttle_mode = -1;
			}
		}

	}
	else if (KEYMOD_CONTROL (kstate))
	{
		if (eng_status != 2)
		{
			RESETKEY (kstate, OAPI_KEY_ADD);
			RESETKEY (kstate, OAPI_KEY_NUMPAD1);
			RESETKEY (kstate, OAPI_KEY_NUMPAD3);
		}

		if ((auto_alt_status + auto_hvr_status) > 0)
		{
			if (KEYDOWN (kstate, OAPI_KEY_ADD)) 
			{
			RESETKEY (kstate, OAPI_KEY_ADD);
			throttle_mode = 1;
			}

		if (KEYDOWN (kstate, OAPI_KEY_SUBTRACT)) 
			{
			RESETKEY (kstate, OAPI_KEY_SUBTRACT);
			throttle_mode = -1;
			}
		}

	}
	else
	{
		if (eng_status != 2)
		{
			RESETKEY (kstate, OAPI_KEY_ADD);
			RESETKEY (kstate, OAPI_KEY_NUMPAD1);
			RESETKEY (kstate, OAPI_KEY_NUMPAD3);
			RESETKEY (kstate, OAPI_KEY_NUMPAD5);
		}

		/////// pitch
		if (KEYDOWN (kstate, OAPI_KEY_NUMPAD2)) 
		{
		RESETKEY (kstate, OAPI_KEY_NUMPAD2);
		pitch_mode = 1;
		}

		if (KEYDOWN (kstate, OAPI_KEY_NUMPAD8)) 
		{
		RESETKEY (kstate, OAPI_KEY_NUMPAD8);
		pitch_mode = -1;
		}

		//// bank
		if (KEYDOWN (kstate, OAPI_KEY_NUMPAD4)) 
		{
		RESETKEY (kstate, OAPI_KEY_NUMPAD4);
		bank_mode = 1;
		}

		if (KEYDOWN (kstate, OAPI_KEY_NUMPAD6)) 
		{
		RESETKEY (kstate, OAPI_KEY_NUMPAD6);
		bank_mode = -1;
		}

		//// pitch trim
		if (KEYDOWN (kstate, OAPI_KEY_N)) 
		{
		RESETKEY (kstate, OAPI_KEY_N);
		rotor_trim = rotor_trim + step_time;
		if (rotor_trim > 1)rotor_trim = 1;
		}

		if (KEYDOWN (kstate, OAPI_KEY_M)) 
		{
		RESETKEY (kstate, OAPI_KEY_M);
		rotor_trim = rotor_trim - step_time;
		if (rotor_trim < 0)rotor_trim = 0;
		}

	}

	return 0;
}


void as350::clbkPostCreation()
{
	if (eng_status == 2) 
	{
		SetAnimation(anim_rtr1_blds_scale, 1);
		SetAnimation(anim_rtr1_trns_scale, 0);
		SetAnimation(anim_rtr2_blds_scale, 1);
		SetAnimation(anim_rtr2_trns_scale, 0);
	}
	else
	{
		SetAnimation(anim_rtr1_blds_scale, 0);
		SetAnimation(anim_rtr1_trns_scale, 1);
		SetAnimation(anim_rtr2_blds_scale, 0);
		SetAnimation(anim_rtr2_trns_scale, 1);
	}

	/// lights
	
		if (light_status == 0)
		{
			SetAnimation(anim_panel1, 0);
			SetAnimation(anim_panel2, 1);
			beac0.active = false;
			beac1.active = false;
			beac2.active = false;
			beac3.active = false;
		}
		else if (light_status == 1) 
		{
			SetAnimation(anim_panel1, 1);
			SetAnimation(anim_panel2, 0);
			beac0.active = true;
			beac1.active = true;
			beac2.active = true;
			beac3.active = true;
		}
	////////////////////
	hGref = GetGravityRef();

	///////////// sound
	MyID=ConnectToOrbiterSoundDLL(GetHandle());

	ReplaceStockSound(MyID,"Sound\\_CustomVesselsSounds\\as350\\helico.wav",REPLACE_MAIN_THRUST);
	RequestLoadVesselWave(MyID,STRT,"Sound\\_CustomVesselsSounds\\as350\\demarragehelico.wav",BOTHVIEW_FADED_FAR);
	RequestLoadVesselWave(MyID,MOTR,"Sound\\_CustomVesselsSounds\\as350\\helico.wav",BOTHVIEW_FADED_FAR);
	SoundOptionOnOff(MyID,PLAYRADIOATC,FALSE);
	SoundOptionOnOff(MyID,PLAYCOUNTDOWNWHENTAKEOFF,FALSE);
	SoundOptionOnOff(MyID,PLAYLANDINGANDGROUNDSOUND,FALSE);
	SoundOptionOnOff(MyID,PLAYCABINAIRCONDITIONING,FALSE);
	SoundOptionOnOff(MyID,PLAYCABINRANDOMAMBIANCE,FALSE);
	SoundOptionOnOff(MyID,DISPLAYTIMER,FALSE);
	SoundOptionOnOff(MyID,PLAYDOCKLANDCLEARANCE,FALSE);
	SoundOptionOnOff(MyID,PLAYATTITUDETHRUST,FALSE);

	SetAttitudeMode (RCS_ROT);

}



void as350::clbkPreStep(double SimT, double SimDT, double MJD)
{
	step_time = SimDT;

	// get joystick input
	js_pu = GetManualControlLevel(THGROUP_ATT_PITCHUP,MANCTRL_ANYMODE,MANCTRL_JOYSTICK);
	js_pd = GetManualControlLevel(THGROUP_ATT_PITCHDOWN,MANCTRL_ANYMODE,MANCTRL_JOYSTICK);
	js_br = GetManualControlLevel(THGROUP_ATT_BANKRIGHT,MANCTRL_ANYMODE,MANCTRL_JOYSTICK);
	js_bl = GetManualControlLevel(THGROUP_ATT_BANKLEFT,MANCTRL_ANYMODE,MANCTRL_JOYSTICK);
	js_yl = GetManualControlLevel(THGROUP_ATT_YAWLEFT,MANCTRL_ANYMODE,MANCTRL_JOYSTICK);
	js_yr = GetManualControlLevel(THGROUP_ATT_YAWRIGHT,MANCTRL_ANYMODE,MANCTRL_JOYSTICK);
	//sprintf(oapiDebugString(), "js_pu %0.2f js_pd %0.2f js_br %0.2f js_bl %0.2f js_yl %0.2f js_yr %0.2f", js_pu, js_pd, js_br, js_bl, js_yl, js_yr);
	js_pitch_level = js_pu - js_pd;
	js_bank_level = js_bl - js_br;
	js_yaw_level = js_yr - js_yl;

	// set  thrust dir 
	SetMainTh(); 

	//////// do micro throttle
	if((auto_alt_status + auto_hvr_status) == 0)
	{
	if(throttle_mode != 0) 
		{
		dthrottle = (throttle_mode * step_time * 0.01);
		double throt_lvl = GetThrusterLevel(th_main[0]) + dthrottle;
		SetThrusterLevel(th_main[0], throt_lvl);
		}
	else dthrottle = 0;
	}
	
	

	///// for hud infos
	VECTOR3 grnd_spd_vect = _V(0,0,0);
	GetGroundspeedVector(FRAME_HORIZON, grnd_spd_vect);
	ground_spd = length(_V(grnd_spd_vect.x,0,grnd_spd_vect.z));
	if(ground_spd > 0)ground_hdg = atan2(grnd_spd_vect.x,grnd_spd_vect.z);
	else ground_hdg = 0;
	if(ground_hdg < 0)ground_hdg = ground_hdg + PI2;

	///////// for hud infos
	DoTgtInfo();

	///// auto functions
	if(auto_alt_status == 1)DoHoldAlt();
	else if(auto_hvr_status == 1)DoHoldPos();

	///// reset for next frame
	pitch_mode = 0;
	bank_mode = 0;
	if (throttle_mode == 0)dthrottle = 0;
	throttle_mode = 0;

	//////// sound ////////////////////////////////////
	if (eng_status == 1)
	{
	if (IsPlaying(MyID,STRT) == false) PlayVesselWave(MyID, STRT, 0, 225, 0);
	}
	else if (eng_status == 2)
	{
	if (IsPlaying(MyID,STRT) == true) StopVesselWave(MyID, STRT);
	double thr_lev = GetThrusterLevel(th_main[0]);
	double motr_level = (0.6 - thr_lev)*2*200;
	if(motr_level < 50) motr_level = 50;
	else if(motr_level > 200) motr_level = 200;
	if(thr_lev < 0.5) PlayVesselWave(MyID, MOTR, LOOP, motr_level, 33075);
	else StopVesselWave(MyID, MOTR);
	}
	else if (eng_status == 3)
	{
		StopVesselWave(MyID, STRT);
		double freq = 5075+(drot*28000);
		double vol = 50+(drot*150);
		PlayVesselWave(MyID, MOTR, LOOP, vol, freq);
	}
	else
	{
		StopVesselWave(MyID, STRT);
		StopVesselWave(MyID, MOTR);
	}


	///// anims ///////////////////////////

	SetAnimation(anim_rtr_tilt, rotor_trim);

	if (eng_status == 0)
	{
		SetAnimation(anim_rtr1_blds_scale, 0);
		SetAnimation(anim_rtr1_trns_scale, 1);
		SetAnimation(anim_rtr2_blds_scale, 0);
		SetAnimation(anim_rtr2_trns_scale, 1);
		SetThrusterGroupLevel(THGROUP_MAIN, 0);
	}
	else if (eng_status == 1)
	{
		double max_rps = 6;
		drot = drot + (step_time * 0.021);
		if(drot > 1)drot = 1;
		double abs_drot = drot * max_rps;
		
		if(abs_drot == max_rps)eng_status = 2;
		
		rtr1_rot_proc = rtr1_rot_proc + (abs_drot * step_time);
		int rtr1_rot_proc_int = (int)(rtr1_rot_proc);
		rtr1_rot_proc = rtr1_rot_proc - rtr1_rot_proc_int; // always between 0...1

		SetAnimation(anim_rtr1_rotb, rtr1_rot_proc);
		SetAnimation(anim_rtr_axle, rtr1_rot_proc);
		SetAnimation(anim_rtr1_blds_scale, 0);
		SetAnimation(anim_rtr1_trns_scale, 1);

		double rtr2_rot_proc = 5*rtr1_rot_proc;
		int rtr2_rot_proc_int = (int) rtr2_rot_proc;
		rtr2_rot_proc = rtr2_rot_proc - rtr2_rot_proc_int;

		SetAnimation(anim_rtr2_rotb, rtr1_rot_proc);
		SetAnimation(anim_rtr2_blds_scale, 0);
		SetAnimation(anim_rtr2_trns_scale, 1);
		//sprintf(oapiDebugString(), "drot %0.3f rtr1_rot_proc %0.3f ", drot, rtr1_rot_proc);
		SetThrusterGroupLevel(THGROUP_MAIN, 0);
	}
	else if (eng_status == 2)
	{
		double max_rps = 6;
		double abs_drot = step_time * max_rps;
		
		rtr1_rot_proc = rtr1_rot_proc + abs_drot;
		int rtr1_rot_proc_int = (int)(rtr1_rot_proc);
		rtr1_rot_proc = rtr1_rot_proc - rtr1_rot_proc_int; // always between 0...1

		SetAnimation(anim_rtr1_rott, rtr1_rot_proc);
		SetAnimation(anim_rtr_axle, rtr1_rot_proc);
		SetAnimation(anim_rtr1_blds_scale, 1);
		SetAnimation(anim_rtr1_trns_scale, 0);

		double rtr2_rot_proc = 5*rtr1_rot_proc;
		int rtr2_rot_proc_int = (int) rtr2_rot_proc;
		rtr2_rot_proc = rtr2_rot_proc - rtr2_rot_proc_int;

		SetAnimation(anim_rtr2_rott, rtr1_rot_proc);
		SetAnimation(anim_rtr2_blds_scale, 1);
		SetAnimation(anim_rtr2_trns_scale, 0);
	}
	else if (eng_status == 3)
	{
		double max_rps = 6;
		drot = drot - (step_time * 0.025);
		if(drot < 0)drot = 0;
		double abs_drot = drot * max_rps;
		
		if(abs_drot == 0)eng_status = 0;
		
		rtr1_rot_proc = rtr1_rot_proc + (abs_drot * step_time);
		int rtr1_rot_proc_int = (int)(rtr1_rot_proc);
		rtr1_rot_proc = rtr1_rot_proc - rtr1_rot_proc_int; // always between 0...1
		if(rtr1_rot_proc < 0)rtr1_rot_proc = rtr1_rot_proc + 1;// always between 0...1

		SetAnimation(anim_rtr1_rotb, rtr1_rot_proc);
		SetAnimation(anim_rtr_axle, rtr1_rot_proc);
		SetAnimation(anim_rtr1_blds_scale, 0);
		SetAnimation(anim_rtr1_trns_scale, 1);

		double rtr2_rot_proc = 5*rtr1_rot_proc;
		int rtr2_rot_proc_int = (int) rtr2_rot_proc;
		rtr2_rot_proc = rtr2_rot_proc - rtr2_rot_proc_int;

		SetAnimation(anim_rtr2_rotb, rtr1_rot_proc);
		SetAnimation(anim_rtr2_blds_scale, 0);
		SetAnimation(anim_rtr2_trns_scale, 1);
		SetThrusterGroupLevel(THGROUP_MAIN, 0);
	}

	///// hook
	if(hook_status == 0)
	{
		SetAnimation(anim_hook, 0);
		if(grapple_status == 1)grapple_status = 0;
		else if(grapple_status == 3)grapple_status = 0;
	}
	else if(hook_status == 1)
	{
		hook_proc = hook_proc + (0.25 * step_time);
		if(hook_proc > 1)
		{
			hook_status = 2;
			hook_proc = 1;
		}
		SetAnimation(anim_hook, hook_proc);
	}
	else if(hook_status == 2)
	{
		SetAnimation(anim_hook, 1);
		if(grapple_status == 0)grapple_status = 1;
	}
	else if(hook_status == 3)
	{
		hook_proc = hook_proc - (0.25 * step_time);
		if(hook_proc < 0)
		{
			hook_status = 0;
			hook_proc = 0;
		}
		SetAnimation(anim_hook, hook_proc);
	}

	///// set att point
	SetHookPos();

	//////////////////// if deployed, check for genesis grapple
	if(grapple_status == 1)CheckForGrapple();

	if (boom_sw == 0 && GetAltitude(ALTMODE_GROUND) < 1.5)
	{
		VECTOR3 ves_ypos_hrz;
		HorizonRot(_V(0,1,0), ves_ypos_hrz);
		double vert_angle_cos = dotp(_V(0,1,0),ves_ypos_hrz);
		if (vert_angle_cos < 0.25) boom_sw = 1;

		VECTOR3 grnd_spd;
		GetGroundspeedVector(FRAME_HORIZON, grnd_spd);
		if (grnd_spd.y < -50)boom_sw = 1;
	}
	if (boom_sw > 0 && boom_sw < 2) DoBoom();

	if (GetPropellantMass(hProp) == 0)boom_sw = 2;

	
}


void as350::SetMainTh (void)
{
	double stp = GetAtmPressure ();
	double alt_factor = pow((stp / 101400),0.33333);
	double new_max_th = MAIN_TH*alt_factor;
	double new_isp = MAIN_ISP*alt_factor;

	SetThrusterMax0(th_main[0], new_max_th);
	SetThrusterIsp(th_main[0], new_isp);

	double max_pitch_angle = (20*RAD);
	double max_bank_angle = (20*RAD);
	VECTOR3 th_axis = _V(0,1,0);
	VECTOR3 pitch_axis = _V(-1,0,0);
	VECTOR3 bank_axis = _V(0,0,1);

	double rotor_trim_angle =-rotor_trim * (10*RAD);
	th_axis = RotMat(th_axis, pitch_axis, rotor_trim_angle);

	if (pitch_mode == 0) // return to centre
	{
		if (pitch_level < 0) 
			{
			pitch_level = pitch_level + step_time;
			if (pitch_level > 0)pitch_level = 0;
			}
		else if (pitch_level > 0) 
			{
			pitch_level = pitch_level - step_time;
			if (pitch_level < 0)pitch_level = 0;
			}
	}
	else // pitch control active
	{
		pitch_level = pitch_level + (pitch_mode * step_time);
		if (pitch_level > 1)pitch_level = 1;
		else if (pitch_level < -1)pitch_level = -1;
	}

	double pitch_angle = max_pitch_angle * (pitch_level + js_pitch_level);
	th_axis = RotMat(th_axis, pitch_axis, pitch_angle);

	/////////////////////////

	if (bank_mode == 0)
	{
		if (bank_level < 0) 
			{
			bank_level = bank_level + step_time;
			if (bank_level > 0)bank_level = 0;
			}
		else if (bank_level > 0) 
			{
			bank_level = bank_level - step_time;
			if (bank_level < 0)bank_level = 0;
			}
	}
	else 
	{
		bank_level = bank_level + (bank_mode * step_time);
		if (bank_level > 1)bank_level = 1;
		else if (bank_level < -1)bank_level = -1;
	}
	bank_axis = _V(0, -th_axis.z, th_axis.y);

	double bank_angle = max_bank_angle * (bank_level + js_bank_level);
	th_axis = RotMat(th_axis, bank_axis, bank_angle);

	SetThrusterDir(th_main[0], th_axis);

	//sprintf(oapiDebugString(), "th_axis %0.3f %0.3f %0.3f rotor_trim_angle %0.3f", th_axis.x,th_axis.y,th_axis.z,rotor_trim_angle*DEG);
}

VECTOR3 as350::RotMat(VECTOR3 CURR_VECT,VECTOR3 ROT_AXIS,double rot_angle)
{
	double ct = cos(rot_angle);
	double st = sin(rot_angle);

	double ux = ROT_AXIS.x;
	double uy = ROT_AXIS.y;
	double uz = ROT_AXIS.z;
	double ux2 = ux*ux;
	double uy2 = uy*uy;
	double uz2 = uz*uz;

	MATRIX3 RG;

	RG.m11 = ux2+(1-ux2)*ct;
	RG.m12 = ux*uy*(1-ct)-uz*st;
	RG.m13 = ux*uz*(1-ct)+uy*st;
	RG.m21 = ux*uy*(1-ct)+uz*st;
	RG.m22 = uy2+(1-uy2)*ct;
	RG.m23 = uy*uz*(1-ct)-ux*st;
	RG.m31 = ux*uz*(1-ct)-uy*st;
	RG.m32 = uy*uz*(1-ct)+ux*st;
	RG.m33 = uz2+(1-uz2)*ct;

	VECTOR3 result_vector = mul(RG, CURR_VECT);
	return result_vector;

}
void as350::CheckForGrapple(void)
{
	VECTOR3 my_pos_global = _V(0,0,0);
	Local2Global(_V(0,-5.707, 3.344),my_pos_global);
	VECTOR3 tgt_pos_global = _V(0,0,0);

	DWORD ves_num = oapiGetVesselCount();
	for (int i = 0; i < (int)(ves_num); i++)
	{
		
		OBJHANDLE hTgt = oapiGetVesselByIndex(i);
		if (hTgt == GetHandle()) continue; // not ourself

		oapiGetGlobalPos(hTgt, &tgt_pos_global);
		double tgt_dist = dist(my_pos_global, tgt_pos_global);
		if (tgt_dist > 50) continue; // out of range
		

		VESSEL *v = oapiGetVesselInterface (hTgt);
		DWORD att_num = v->AttachmentCount(true);
		if(att_num == 0) continue;// no att points
		

		for (int j = 0; j < (int)(att_num);j++)
		{
		ATTACHMENTHANDLE hTgtAtt = v->GetAttachmentHandle(true, j);
		char cbuf[256];
		strcpy(cbuf, v->GetAttachmentId (hTgtAtt));
		
		if (!strnicmp (cbuf, "GEN", 3))
			{
				VECTOR3 tgt_att_pos;
				VECTOR3 tgt_att_dir;
				VECTOR3 tgt_att_rot;
				v->GetAttachmentParams(hTgtAtt, tgt_att_pos,tgt_att_dir,tgt_att_rot);
				VECTOR3 tgt_att_pos_glob;
				v->Local2Global(tgt_att_pos, tgt_att_pos_glob);
				if (dist(tgt_att_pos_glob,my_pos_global) < 5)
				{
					if(AttachChild(hTgt, hatp, hTgtAtt) == true)grapple_status = 2;
				}
				return;
			}
		}
		if(i == (ves_num - 1)) return;
	}


}


void as350::SetHookPos(void)
{
// hook att //////
	double pole_length = 3.713;
	double extension = 1.5 * hook_proc;
	pole_length = pole_length + extension;

	VECTOR3 pole_vector = _V(0,0, pole_length);
	double pole_rot_angle = 105*RAD*hook_proc;
	pole_vector = RotMat(pole_vector, _V(1,0,0), pole_rot_angle);
	VECTOR3 origin = _V(0,-1.193,0.737);
	VECTOR3 hook_pos = origin + pole_vector;
	SetAttachmentParams(hatp, hook_pos, _V(0,0,1), _V(0,1,0));
	
	// //
}

void as350::DoTgtInfo(void)
{
	if (!strnicmp (tgt_name, "NULL", 4))
	{
	tgt_brg = 0;
	tgt_dist = 0;
	tgt_delta_alt = 0;
	return;
	}
	else
	{
	OBJHANDLE htgt = oapiGetObjectByName(tgt_name);
	if (htgt == NULL)htgt = oapiGetBaseByName(hGref, tgt_name);
	if (htgt == NULL)
		{
		strcpy(tgt_name, "NULL");
		tgt_brg = 0;
		tgt_dist = 0;
		tgt_delta_alt = 0;
		return;
		}
	else 
		{
		if (grapple_status == 2)
			{
			if (htgt == GetAttachmentStatus(hatp))
				{
				tgt_brg = 0;
				tgt_dist = 0;
				tgt_delta_alt = 0;
				return;
				}
			}

			VECTOR3 my_pos_glob = _V(0,0,0);
			VECTOR3 tgt_pos_glob = _V(0,0,0);
			GetRelativePos(hGref, my_pos_glob);
			oapiGetRelativePos(htgt, hGref, &tgt_pos_glob);

			//tgt_dist = dist(my_pos_glob, tgt_pos_glob);

			double tgt_alt = length(tgt_pos_glob) - 6371010;
			tgt_delta_alt = tgt_alt - GetAltitude(ALTMODE_MEANRAD);

			oapiGetGlobalPos(htgt, &tgt_pos_glob);
			VECTOR3 tgt_pos_loc = _V(0,0,0);
			Global2Local(tgt_pos_glob, tgt_pos_loc);

			VECTOR3 tgt_pos_hrz = _V(0,0,0);
			HorizonRot(tgt_pos_loc, tgt_pos_hrz);

			tgt_dist = length(_V(tgt_pos_hrz.x, 0, tgt_pos_hrz.z));
			tgt_brg = atan2(tgt_pos_hrz.x, tgt_pos_hrz.z);
			if(tgt_brg < 0)tgt_brg = tgt_brg + PI2;

		}
	}

}


bool as350::clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp)
{
	// draw the default HUD
	VESSEL3::clbkDrawHUD (mode, hps, skp);

	int s = hps->H;
	double d = (s*0.00130208);
	int sw = ((hps->W));
	int lw =  (int)(16*sw/1024);
	int lwoffset = sw - (18*lw);
	int hlw = (int) (lw/2);


	int roxl = 0;
	int royl = 0;

	double ds = s;
	double dsw = sw;
	double sc_ratio = ds/dsw;


	if (sc_ratio < 0.7284)
	{
	roxl = (lw*10);
	royl = (int) (-88*d);
	}



	int wd = (int)(152*d);
	int wc = (int)(168*d);
	int w0 = (int)(184*d);
	int w1 = (int) (200*d);
	int w2 = (int)(216*d);
	int w3 = (int)(232*d);
	int w4 = (int)(248*d);
	int w5 = (int)(264*d);
	int w6 = (int)(280*d);
	int w7 = (int)(296*d);
	int w8 = (int)(312*d);
	int w9 = (int)(328*d);
	int w10 = (int)(344*d);
	int w11 = (int)(360*d);

	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];

	if(hook_status == 0)sprintf(abuf,"G = Deploy Grapple");
	else if(hook_status == 1)sprintf(abuf,"G = Stow Grapple");
	else if(hook_status == 2)
		{
		if(grapple_status == 3)sprintf(abuf,"G = Reset Grapple");
		else sprintf(abuf,"G = Stow Grapple");
		}
	else if(hook_status == 3)
		{
		if(grapple_status == 3)sprintf(abuf,"G = Cancel Reset Grapple");
		else sprintf(abuf,"G = Stow Grapple");
		}
	skp->Text ((10+roxl), (wc+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"J = Release Grapple");
	skp->Text ((10+roxl), (w0+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"L = Night Lights On/Off");
	skp->Text ((10+roxl), (w1+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"N/M = Rotor Tilt: %0.1f", (rotor_trim * 10));
	skp->Text ((10+roxl), (w2+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"Insert/Delete = Elevator Trim");
	skp->Text ((10+roxl), (w3+royl), abuf,strlen(abuf));
	}

	{
	if(auto_hvr_status == 0) skp->SetTextColor(0x0066FF66);
	else skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"V = AutoHover On/Off");
	skp->Text ((10+roxl), (w4+royl), abuf,strlen(abuf));
	}

	{
	if(auto_alt_status == 0) skp->SetTextColor(0x0066FF66);
	else skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"B = Hold Alt. On/Off");
	skp->Text ((10+roxl), (w5+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"Ground Speed %0.0f kts  Heading %0.1f",(1.94384*ground_spd), (ground_hdg*DEG));
	skp->Text ((10+roxl), (w6+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"P = Select Target: %s BRG %0.0f",tgt_name, DEG*tgt_brg);
	skp->Text ((10+roxl), (w8+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"Tgt. DIST %0.1f km  REL.ALT %0.0f m", tgt_dist*0.001, tgt_delta_alt);
	skp->Text ((10+roxl), (w9+royl), abuf,strlen(abuf));
	}

	

	return true;
}

void as350::DoHoldPos(void)
{
	if(rotor_trim > 0)
	{	
	rotor_trim = rotor_trim - (step_time);
	if(rotor_trim < 0) rotor_trim = 0;
	}

	VECTOR3 grd_spd_vec_hrz;
	GetGroundspeedVector(FRAME_HORIZON, grd_spd_vec_hrz);

	VECTOR3 weight_vec;
	GetWeightVector(weight_vec);
	double weight = length(weight_vec);
	weight_vec = _V(0,-weight,0);

	VECTOR3 drag_vector;
	GetDragVector(drag_vector);

	double stop_time = ground_spd/5;
	if(stop_time < 2)stop_time = 2;
	double mass = GetMass();
	double angle_limit = 10*RAD;

	double req_acc_x = -grd_spd_vec_hrz.x/stop_time;
	double req_acc_y = -grd_spd_vec_hrz.y/stop_time;
	double req_acc_z = -grd_spd_vec_hrz.z/stop_time;

	VECTOR3 req_force_x = _V(req_acc_x*mass,0,0);
	VECTOR3 req_force_y = _V(0,req_acc_y*mass,0);
	VECTOR3 req_force_z = _V(0,0,req_acc_z*mass);

	VECTOR3 req_thrust_hrz = req_force_x + req_force_y + req_force_z - weight_vec;// - drag_vector; 

	double max_th = GetThrusterMax(th_main[0]);

	double th_level = length(req_thrust_hrz)/max_th;
	

	VECTOR3 req_thrust_loc;
	HorizonInvRot(req_thrust_hrz, req_thrust_loc);
	normalise(req_thrust_loc);

	double req_angle = acos(dotp(_V(0,1,0), req_thrust_loc));
	if (req_angle > angle_limit)
	{
		VECTOR3 th_rot_axis = crossp(_V(0,1,0),req_thrust_loc);
		normalise(th_rot_axis);
		req_thrust_loc = RotMat(_V(0,1,0), th_rot_axis, angle_limit);
		normalise(req_thrust_loc);
		//sprintf(oapiDebugString(), "rotaxis %0.3f %0.3f %0.3f", th_rot_axis.x, th_rot_axis.y, th_rot_axis.z);
	}

	////////////////////// user input //////////////

	//////////// pitch n bank
	VECTOR3 pitch_axis = _V(-1,0,0);
	VECTOR3 bank_axis = _V(0,0,1);

	///// pitch
	if (pitch_mode == 0) // return to centre
	{
		if (pitch_level < 0) 
			{
			pitch_level = pitch_level + step_time;
			if (pitch_level > 0)pitch_level = 0;
			}
		else if (pitch_level > 0) 
			{
			pitch_level = pitch_level - step_time;
			if (pitch_level < 0)pitch_level = 0;
			}
	}
	else // pitch control active
	{
		pitch_level = pitch_level + (pitch_mode * step_time);
		if (pitch_level > 1)pitch_level = 1;
		else if (pitch_level < -1)pitch_level = -1;
	}

	double user_pitch_angle = (js_pitch_level + pitch_level)*(10*RAD);
	req_thrust_loc = RotMat(req_thrust_loc, pitch_axis, user_pitch_angle);

	//// bank

	if (bank_mode == 0)
	{
		if (bank_level < 0) 
			{
			bank_level = bank_level + step_time;
			if (bank_level > 0)bank_level = 0;
			}
		else if (bank_level > 0) 
			{
			bank_level = bank_level - step_time;
			if (bank_level < 0)bank_level = 0;
			}
	}
	else 
	{
		bank_level = bank_level + (bank_mode * step_time);
		if (bank_level > 1)bank_level = 1;
		else if (bank_level < -1)bank_level = -1;
	}

	double user_bank_angle = (js_bank_level + bank_level)*(10*RAD);
	req_thrust_loc = RotMat(req_thrust_loc, bank_axis, user_bank_angle);

	///////////// throttle


	if(throttle_mode != 0) 
	{
		dthrottle = (throttle_mode * step_time * 0.1);
		double throt_lev = GetThrusterLevel(th_main[0]);
		throt_lev = throt_lev + dthrottle;
		if (throt_lev > 1)throt_lev = 1;
		else if(throt_lev < 0)throt_lev = 0;
		SetThrusterLevel(th_main[0],throt_lev);
	}
	else 
	{
		if (th_level > 1)th_level = 1;
		else if(th_level < 0)th_level = 0;
		SetThrusterLevel(th_main[0], th_level);
	}

	SetThrusterDir(th_main[0], req_thrust_loc);

	//// y axis //////////////////////
	VECTOR3 ang_vel;
	GetAngularVel(ang_vel);

	double drot_y = step_time*ang_vel.y;
	double max_y_torque = 3*YAW_TH;
	double max_y_acc = max_y_torque/(mass*6);
	double max_y_dvrot = max_y_acc*step_time;
	double req_level = ang_vel.y/max_y_dvrot;
	if(req_level > 0.5)req_level = 0.5;
	else if(req_level < -0.5)req_level = -0.5;
	if(req_level > 0)
	{
	SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT,0);
	SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT,req_level);
	}
	else
	{
	SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT,-req_level);
	SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT,0);
	}
	

}

void as350::DoHoldAlt(void)
{
	VECTOR3 th_axis_loc;
	GetThrusterDir(th_main[0],th_axis_loc);
	VECTOR3 th_axis_hrz;
	HorizonRot(th_axis_loc, th_axis_hrz);
	normalise(th_axis_hrz);

	VECTOR3 grd_spd_hrz;
	GetGroundspeedVector(FRAME_HORIZON, grd_spd_hrz);
	double stop_time = abs(grd_spd_hrz.y)/5;
	if(stop_time < 0.1)stop_time = 0.1;
	double req_acc = -grd_spd_hrz.y/stop_time;
	double mass = GetMass();
	double vvel_force = mass*req_acc;

	VECTOR3 weight_vect;
	GetWeightVector(weight_vect);
	double weight = length(weight_vect);
	double req_vert_force = weight + vvel_force;

	double req_total_force = req_vert_force/th_axis_hrz.y;

	double th_lev = req_total_force/GetThrusterMax(th_main[0]);

	if(throttle_mode != 0) 
	{
		dthrottle = (throttle_mode * step_time * 0.1);
		double throt_lev = GetThrusterLevel(th_main[0]);
		throt_lev = throt_lev + dthrottle;
		if (throt_lev > 1)throt_lev = 1;
		else if(throt_lev < 0)throt_lev = 0;
		if (th_axis_hrz.y > 0) SetThrusterLevel(th_main[0], throt_lev);
	}
	else 
	{
		if (th_lev > 1)th_lev = 1;
		else if(th_lev < 0)th_lev = 0;
		if (th_axis_hrz.y > 0) SetThrusterLevel(th_main[0], th_lev);
	}


}

void as350::DoBoom(void)
{
//if(boom_sw == 1)
//{
VESSELSTATUS2 vs;
	memset (&vs, 0, sizeof(vs));
	vs.version = 2;
	GetStatusEx(&vs);
	OBJHANDLE hVessel = GetHandle();
	double longitude, latitude, radius;
	double heading;
	oapiGetEquPos(hVessel,&longitude,&latitude,&radius);
	oapiGetHeading(hVessel,&heading);
	vs.vrot.x = 1.43;
	vs.surf_lat = latitude;
	vs.surf_lng = longitude;
	vs.surf_hdg = heading;
	vs.status = 1;
	DefSetStateEx (&vs);

	eng_status = 0;
	drot = 0;
	light_status = 0;
	SetAnimation(anim_panel1, 0);
	SetAnimation(anim_panel2, 1);
	beac0.active = false;
	beac1.active = false;
	beac2.active = false;
	beac3.active = false;
//}
	//SetMeshVisibilityMode(0,MESHVIS_NEVER);
	//SetMeshVisibilityMode(1,MESHVIS_NEVER);

	boom_sw = boom_sw + step_time;
	if (boom_sw < 2)
	{
		SetThrusterLevel(th_boom[0],1);
		SetThrusterLevel(th_boom[1],1);
	}
	else
	{
		boom_sw = 2;
		SetThrusterLevel(th_boom[0],0);
		SetThrusterLevel(th_boom[1],0);
		SetPropellantMass(hProp, 0);
	}
}