// ==============================================================
//                 ORBITER MODULE: genesis_bus
//                  
//
// genesis_bus.cpp
// Control module for genesis_bus vessel class
//
// 
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"

// ==============================================================
// Some vessel parameters
// ==============================================================

const double PA_ISP = 3000;
const double PB_MAXRCSTH = 1;
const double PB_MAXRCSTH20 = 22;
const double GEN_BUS_MASS = 289;
const double GEN_SRC_MASS = 205;
const double SOLP_OPERATING_SPEED = 0.2;
const double TETHER_OPERATING_SPEED = 1;
const double HINGE_OPERATING_SPEED = 0.05;
#define RED RGB(255, 0, 0)
#define GREEN RGB(0, 255, 0)
#define YELLOW RGB(255, 255, 0)

static const DWORD ntdvtx_geardown = 7;
static TOUCHDOWNVTX tdvtx_geardown[ntdvtx_geardown] = {
	{_V(     0, -1,-0.5), 2e3, 1e4, 2, 2},
	{_V(-0.866,0.5,-0.5), 2e3, 1e4, 2, 2},
	{_V( 0.866,0.5,-0.5), 2e3, 1e4, 2, 2},
	{_V(   0,   0.5,   0), 1e5, 1e6, 2},
	{_V(    0,    1,-0.5), 2e3, 1e4, 2},
	{_V(-0.866,-0.5,-0.5), 2e3, 1e4, 2},
	{_V( 0.866,-0.5,-0.5), 2e3, 1e4, 2},
};


// Calculate lift coefficient [Cl] as a function of aoa (angle of attack) over -Pi ... Pi
// Implemented here as a piecewise linear function
double LiftCoeff (double aoa)
{
	const int nlift = 9;
	static const double AOA[nlift] = {-180*RAD,-60*RAD,-30*RAD,-1*RAD,15*RAD,20*RAD,25*RAD,60*RAD,180*RAD};
	static const double CL[nlift]  = {       0,      0,		0,		0,		0,		0,		0,     0,      0};
	static const double SCL[nlift] = {(CL[1]-CL[0])/(AOA[1]-AOA[0]), (CL[2]-CL[1])/(AOA[2]-AOA[1]),
		                              (CL[3]-CL[2])/(AOA[3]-AOA[2]), (CL[4]-CL[3])/(AOA[4]-AOA[3]),
									  (CL[5]-CL[4])/(AOA[5]-AOA[4]), (CL[6]-CL[5])/(AOA[6]-AOA[5]),
									  (CL[7]-CL[6])/(AOA[7]-AOA[6]), (CL[8]-CL[7])/(AOA[8]-AOA[7])};
	int i;
	for (i = 0; i < nlift-1 && AOA[i+1] < aoa; i++);

	return CL[i] + (aoa-AOA[i])*SCL[i];
}


class genesis_bus: public VESSEL3 {
public:
	enum DoorStatus { DOOR_CLOSED, DOOR_OPEN, DOOR_CLOSING, DOOR_OPENING } ant_status, fold_status, hinge_status;
	genesis_bus (OBJHANDLE hVessel, int flightmodel);
	~genesis_bus ();
	
	void clbkSetClassCaps (FILEHANDLE cfg);
	void clbkPreStep(double SimT, double SimDT, double MJD);
	void clbkPostCreation();
	void clbkSaveState (FILEHANDLE scn);
	void clbkLoadStateEx (FILEHANDLE scn, void *status);
	bool clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp);
	int clbkConsumeBufferedKey(DWORD key, bool down, char *kstate);

	private:
	double CalcEmptyMass();
	void DefineAnimations (void);
	void RevertAntenna (void);
	void RevertArray (void);
	void RevertHinge (void);
	void SpawnObject(char* classname, char* ext, VECTOR3 ofs, VECTOR3 vofs);
	void Jettison();
	int status;
	THRUSTER_HANDLE th_rcs[20], th_group[4];
	PROPELLANT_HANDLE rcspr;
	ANIMATIONCOMPONENT_HANDLE Parent;


	UINT anim_ant, anim_fold, anim_hinge;
	double ant_proc, fold_proc, hinge_proc;
};

genesis_bus::genesis_bus (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
    

}

genesis_bus::~genesis_bus ()
{

}


// Define animation sequences for moving parts
// --------------------------------------------------------------
void genesis_bus::DefineAnimations (void)
{
	// 1. tether
	static UINT HiGainAnt1Grp[1] = {19};
	static MGROUP_ROTATE HiGainAnt1 (0, HiGainAnt1Grp, 1, _V(0.43,0.492,-0.022), _V(-0.707107,0.707107,0), (float)(PI*0.4));
	static UINT HiGainAnt2Grp[5] = {1,2,7,8,9};
	static MGROUP_ROTATE HiGainAnt2 (0, HiGainAnt2Grp, 5, _V(0.531,0.593,0.319), _V(-0.707107,0.707107,0), (float)(PI*0.05));
	anim_ant = CreateAnimation (0);
	Parent = AddAnimationComponent (anim_ant, 0, 1, &HiGainAnt1);
	AddAnimationComponent (anim_ant, 0, 1,   &HiGainAnt2, Parent);

	// 2. hinge
	static UINT HingeGrp1[2] = {2,9};
	static MGROUP_ROTATE Hinge1 (0, HingeGrp1, 2, _V( 0.53, 0.619, 0.383), _V(-0.707107,0.707107,0), (float)(PI*1));
	static UINT HingeGrp2[6] = {1,2,3,4,5,6};
	static MGROUP_ROTATE Hinge2 (1, HingeGrp2, 6, _V( 0.53, 0.619, 0.383), _V(-0.707107,0.707107,0), (float)(PI*1));

	static UINT HingeGrp3[2] = {14,15};
	static MGROUP_ROTATE Hinge3 (1, HingeGrp3, 2, _V( -0.442,-0.351,0.506), _V( 0.707107,-0.707107,0), (float)(PI*1));

	static UINT collgrp1[3] = {11,12,13};
	static MGROUP_ROTATE Coll1 (1, collgrp1, 3, _V( 0.196, 0.196, 0), _V(0,0,-1), (float)(PI*0.5));

	static UINT collgrp2[1] = {10};
	static MGROUP_ROTATE Coll2 (1, collgrp2, 1, _V( 0.196, 0.196, 0), _V(0,0,1), (float)(PI*0.5));

	anim_hinge = CreateAnimation (0);
	AddAnimationComponent (anim_hinge, 0, 0.3, &Hinge1);
	AddAnimationComponent (anim_hinge, 0, 0.3, &Hinge2);
	AddAnimationComponent (anim_hinge, 0.35, 0.6, &Hinge3);

	AddAnimationComponent (anim_hinge, 0.65, 1, &Coll1);
	AddAnimationComponent (anim_hinge, 0.65, 1, &Coll2);

	
	// 3. Solar arrays - folding
	anim_fold = CreateAnimation (0);
	static UINT ArrayLFoldGrp[4] = {23,25,27,28};
	static UINT ArrayRFoldGrp[4] = {24,26,29,30};
	
	static MGROUP_ROTATE ArrayLFold1 (0, ArrayLFoldGrp, 4, _V( 1.061, 0, -0.146), _V(0,1,0), (float)(PI*0.63333));
	AddAnimationComponent (anim_fold, 0, 1, &ArrayLFold1);
	static MGROUP_ROTATE ArrayLFold2 (0, ArrayRFoldGrp, 4, _V(-1.061, 0, -0.146), _V(0,-1,0), (float)(PI*0.63333));
	AddAnimationComponent (anim_fold, 0, 1, &ArrayLFold2);

}

void genesis_bus::RevertAntenna (void)
{
	ant_status = ((ant_status == DOOR_CLOSED || ant_status == DOOR_CLOSING) ?
		DOOR_OPENING : DOOR_CLOSING);
}

void genesis_bus::RevertArray (void)
{
	fold_status = ((fold_status == DOOR_CLOSED || fold_status == DOOR_CLOSING) ?
		DOOR_OPENING : DOOR_CLOSING);
}

void genesis_bus::RevertHinge (void)
{
	hinge_status = ((hinge_status == DOOR_CLOSED || hinge_status == DOOR_CLOSING) ?
		DOOR_OPENING : DOOR_CLOSING);
}

// ==============================================================
// genesis_bus class interface
// ==============================================================

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void genesis_bus::clbkSetClassCaps (FILEHANDLE cfg)
{
	ant_proc = 0.0;
	ant_status = DOOR_CLOSED;
	fold_proc = 0.0;
	fold_status = DOOR_CLOSED;
	hinge_proc = 0.0;
	hinge_status = DOOR_CLOSED;
	DefineAnimations ();
	status = 0;
	// physical specs
	SetCW (0.3, 0.3, 0.6, 0.9);
	SetWingAspect (0.7);
	SetWingEffectiveness (2.5);
	SetRotDrag (_V(0.35,0.6,0.6));
	if (GetFlightModel() >= 1) {
		SetPitchMomentScale (1e-4);
		SetBankMomentScale (1e-4);
	}
	SetTrimScale (0.05);
	SetPMI (_V(0.4,0.5,0.6));
	SetCrossSections (_V(3,3,8));
	SetSize (3.0);
	SetMeshVisibilityMode ((AddMesh ("GenesisMission/gen_bus1")), MESHVIS_ALWAYS);
	SetMeshVisibilityMode ((AddMesh ("GenesisMission/gen_bus2")), MESHVIS_ALWAYS);
	SetCameraOffset (_V(0,2,-4));
	SetLiftCoeffFunc (LiftCoeff);
	//SetTouchdownPoints (_V(0,1,-0.5), _V(1,-1,-0.5), _V(-1,-1,-0.5));
	SetTouchdownPoints (tdvtx_geardown,ntdvtx_geardown);
	//SetDockParams (_V(0,0,0), _V(0,1,0), _V(0,0,-1));

	// propellant resources
	const double PA_FUELMASS = 142;
	double rcsmass = 142;
	rcspr = CreatePropellantResource (PA_FUELMASS, rcsmass);
	// ***************** thruster definitions *******************

	// lin z+
	th_rcs[ 0] = CreateThruster (_V( 0.887, -0.355, -0.47), _V(0, 0, 1), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[ 1] = CreateThruster (_V( -0.887, -0.355, -0.47), _V(0, 0, 1), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[ 2] = CreateThruster (_V( 0.887, 0.355, -0.47), _V(0, 0, 1), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[ 3] = CreateThruster (_V( -0.887, 0.355, -0.47), _V(0, 0, 1), PB_MAXRCSTH, rcspr, PA_ISP);
	// yaw left
	th_rcs[ 4] = CreateThruster (_V( 0.887, 0.355, -0.47), _V(0, 0, 1), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[ 5] = CreateThruster (_V( 0.887, -0.355, -0.47), _V(0, 0, 1), PB_MAXRCSTH, rcspr, PA_ISP);
	// yaw right
	th_rcs[ 6] = CreateThruster (_V( -0.887, 0.355, -0.47), _V(0, 0, 1), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[ 7] = CreateThruster (_V( -0.887, -0.355, -0.47), _V(0, 0, 1), PB_MAXRCSTH, rcspr, PA_ISP);
	
	// pitch up
	th_rcs[ 8] = CreateThruster (_V( 0.852, 0.444, -0.394), _V(0, -1, 0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[ 9] = CreateThruster (_V( -0.852, 0.444, -0.394), _V(0, -1, 0), PB_MAXRCSTH, rcspr, PA_ISP);

	// pitch down
	th_rcs[10] = CreateThruster (_V( 0.852, -0.444, -0.394), _V(0, 1, 0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[11] = CreateThruster (_V( -0.852, -0.444, -0.394), _V(0, 1, 0), PB_MAXRCSTH, rcspr, PA_ISP);

	// bank left
	th_rcs[12] = CreateThruster (_V( 0.852, -0.444, -0.394), _V(0, 1, 0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[13] = CreateThruster (_V( -0.852, 0.444, -0.394), _V(0, -1, 0), PB_MAXRCSTH, rcspr, PA_ISP);

	// bank right
	th_rcs[14] = CreateThruster (_V( -0.852, -0.444, -0.394), _V(0, 1, 0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[15] = CreateThruster (_V( 0.852, 0.444, -0.394), _V(0, -1, 0), PB_MAXRCSTH, rcspr, PA_ISP);

	// main
	th_rcs[16] = CreateThruster (_V( 0.817, -0.043, -0.45), _V(0, 0, 1), PB_MAXRCSTH20, rcspr, PA_ISP);
	th_rcs[17] = CreateThruster (_V( -0.817, -0.043, -0.45), _V(0, 0, 1), PB_MAXRCSTH20, rcspr, PA_ISP);
	th_rcs[18] = CreateThruster (_V( 0.817, 0.043, -0.45), _V(0, 0, 1), PB_MAXRCSTH20, rcspr, PA_ISP);
	th_rcs[19] = CreateThruster (_V( -0.817, 0.043, -0.45), _V(0, 0, 1), PB_MAXRCSTH20, rcspr, PA_ISP);
	

	th_group[0] = th_rcs[ 8];
	th_group[1] = th_rcs[ 9];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHUP);

	th_group[0] = th_rcs[10];
	th_group[1] = th_rcs[11];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHDOWN);

	th_group[0] = th_rcs[12];
	th_group[1] = th_rcs[13];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKLEFT);

	th_group[0] = th_rcs[14];
	th_group[1] = th_rcs[15];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKRIGHT);

	th_group[0] = th_rcs[ 4];
	th_group[1] = th_rcs[ 5];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWLEFT);

	th_group[0] = th_rcs[ 6];
	th_group[1] = th_rcs[ 7];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWRIGHT);


	th_group[0] = th_rcs[ 0];
	th_group[1] = th_rcs[ 1];
	th_group[2] = th_rcs[ 2];
	th_group[3] = th_rcs[ 3];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_FORWARD);

	th_group[0] = th_rcs[16];
	th_group[1] = th_rcs[17];
	th_group[2] = th_rcs[18];
	th_group[3] = th_rcs[19];
	CreateThrusterGroup (th_group, 4, THGROUP_MAIN);

	SURFHANDLE exhaust1 = oapiRegisterExhaustTexture ("exhaust_atrcs");

	for(int i=0;i<16;i++)
	{
	AddExhaust (th_rcs[i], 0.6, 0.03,(exhaust1));
	}
	for(int i=16;i<20;i++)
	{
	AddExhaust (th_rcs[i], 2, 0.08,(exhaust1));
	}

}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new genesis_bus (hvessel, flightmodel);
}
// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (genesis_bus*)vessel;
}
// ==============================================================
// Overloaded clbkPostCreation callback functions
// ==============================================================

void genesis_bus::clbkPostCreation ()
{
	double rcsprmass = 142;
	if (GetPropellantMass(rcspr)< 0.01 && oapiGetSimTime ()> 0.1)SetPropellantMass(rcspr, rcsprmass);
	if (status > 0) DelMesh(1);
}
void genesis_bus::clbkSaveState (FILEHANDLE scn)
{
	char cbuf[256];

	// default vessel parameters
	VESSEL3::clbkSaveState (scn);

	// custom parameters
	oapiWriteScenario_int (scn, "CONFIGURATION", status);
	sprintf (cbuf, "%d %0.4f", ant_status, ant_proc);
	oapiWriteScenario_string (scn, "TETHER", cbuf);
	sprintf (cbuf, "%d %0.4f", fold_status, fold_proc);
	oapiWriteScenario_string (scn, "SOLP", cbuf);
	sprintf (cbuf, "%d %0.4f", hinge_status, hinge_proc);
	oapiWriteScenario_string (scn, "BACKSHELL", cbuf);
	//sprintf (cbuf, "%0.1f %0.1f %0.1f", testpos.x, testpos.y, testpos.z);
	//oapiWriteScenario_string (scn, "TESTPOS", cbuf);
	//sprintf (cbuf, "%0.1f %0.1f %0.1f", testpos2.x, testpos2.y, testpos2.z);
	//oapiWriteScenario_string (scn, "TESTPOS2", cbuf);
}
void genesis_bus::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
    char *line;
	while (oapiReadScenario_nextline (scn, line)) {
        if (!strnicmp (line, "CONFIGURATION", 13)) {
            sscanf (line+13, "%d", &status);
		} else if (!strnicmp (line, "TETHER", 6)) {
			sscanf (line+6, "%d%lf", &ant_status, &ant_proc);
		} else if (!strnicmp (line, "SOLP", 4)) {
			sscanf (line+4, "%d%lf", &fold_status, &fold_proc);
		} else if (!strnicmp (line, "BACKSHELL", 9)) {
			sscanf (line+9, "%d%lf", &hinge_status, &hinge_proc);
		}else {
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
   }
	SetAnimation (anim_ant, ant_proc);
	SetAnimation (anim_fold, fold_proc);
	SetAnimation (anim_hinge, hinge_proc);
}
int genesis_bus::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate) {
  if (!down) return 0; // only process keydown events

  if (KEYMOD_SHIFT (kstate)) {

  } else { // unmodified keys
    switch (key) {
    case OAPI_KEY_J:  // jettison
		if(hinge_status == 0 && fold_status == 1 && status < 1)
		{
        Jettison();
		RevertAntenna();
		}
        return 1;
	case OAPI_KEY_K: // open cover
		if(status < 1 && fold_status == 1)RevertHinge();
		return 1;
	case OAPI_KEY_G: // open solar arrays
		if(fold_status == 0)RevertArray();
		return 1;
	}
  }
  return 0;
}
void genesis_bus::SpawnObject(char* classname, char* ext, VECTOR3 ofs, VECTOR3 vofs) {

	VECTOR3 pos;
	VECTOR3 vel;
	GlobalRot (_V(-1,0,0), pos);
	GlobalRot (_V(0,0,-1), vel);
	VECTOR3 yup;
	VECTOR3 yupn;
	VECTOR3 zforward;
	VECTOR3 xright;
	yupn = crossp(vel, pos);
	double yupnlength;
	yupnlength = length(yupn);
	yup = _V(yupn.x/yupnlength, yupn.y/yupnlength, yupn.z/yupnlength);
	xright = crossp (yup, vel);
	zforward = vel;
	
	double gamma = atan2( xright.x, yup.x);
	double beta = -1*(asin(zforward.x));
	double alpha = atan2(zforward.y, zforward.z);

  VESSELSTATUS vs;
  char name[256];
  GetStatus(vs);
  VECTOR3 rvel;
  Local2Rel (ofs, vs.rpos);
  GlobalRot (vofs, rvel);
  vs.rvel = vs.rvel + rvel;
  vs.arot = _V(alpha, beta, (PI/2)-gamma);
  vs.vrot.x = -vs.vrot.x;
  vs.vrot.z = -vs.vrot.z;
  vs.eng_main = vs.eng_hovr = 0.0;
  vs.status = 0;
  strcpy (name, ext);
  oapiCreateVessel (name, classname, vs);
}
void genesis_bus::Jettison() {
  switch(status + hinge_status) {
    case 0:
      status=1;
      SpawnObject("Vessels/GenesisMission/genesis_src","Genesis_SRC",_V(0,0,0.2796),_V(0,0,0.3));
      break;
  }
}
double genesis_bus::CalcEmptyMass() {
	double EmptyMass=0;
	if (status < 1) {
	EmptyMass+=GEN_SRC_MASS;
	}

	EmptyMass+=GEN_BUS_MASS;
	return EmptyMass;
} 
void genesis_bus::clbkPreStep(double SimT, double SimDT, double MJD) {
	// next block of code
	SetEmptyMass(CalcEmptyMass());

	/// from post step
	if (status > 0) DelMesh(1);

	// Animate tether retract
	if (ant_status >= DOOR_CLOSING) {
		double da = SimDT * TETHER_OPERATING_SPEED;
		if (ant_status == DOOR_CLOSING) {
			if (ant_proc > 0.0) ant_proc = max (0.0, ant_proc-da);
			else                ant_status = DOOR_CLOSED;
		} else {
			if (ant_proc < 1.0) ant_proc = min (1.0, ant_proc+da);
			else                ant_status = DOOR_OPEN;
		}
		SetAnimation (anim_ant, ant_proc);
	}

	// Animate solar arrays
	if (fold_status >= DOOR_CLOSING) {
		double da = SimDT * SOLP_OPERATING_SPEED;
		if (fold_status == DOOR_CLOSING) {
			if (fold_proc > 0.0) fold_proc = max (0.0, fold_proc-da);
			else                 fold_status = DOOR_CLOSED;
		} else {
			if (fold_proc < 1.0) fold_proc = min (1.0, fold_proc+da);
			else                 fold_status = DOOR_OPEN;
		}
		SetAnimation (anim_fold, fold_proc);
	}

	// Animate backshell hinge
	if (hinge_status >= DOOR_CLOSING) {
		double da = SimDT * HINGE_OPERATING_SPEED;
		if (hinge_status == DOOR_CLOSING) {
			if (hinge_proc > 0.0) hinge_proc = max (0.0, hinge_proc-da);
			else                 hinge_status = DOOR_CLOSED;
		} else {
			if (hinge_proc < 1.0) hinge_proc = min (1.0, hinge_proc+da);
			else                 hinge_status = DOOR_OPEN;
		}
		SetAnimation (anim_hinge, hinge_proc);
	}

}

bool genesis_bus::clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp)
{
	// draw the default HUD
	VESSEL3::clbkDrawHUD (mode, hps, skp);

	int s = hps->H;
	double d = (s*0.00130208);
	int sw = ((hps->W));
	int lw =  (int)(16*sw/1024);
	int lwoffset = sw - (18*lw);
	int hlw = (int) (lw/2);


	int roxl = 0;
	int royl = 0;

	double ds = s;
	double dsw = sw;
	double sc_ratio = ds/dsw;


	if (sc_ratio < 0.7284)
	{
	roxl = (lw*10);
	royl = (int) (-88*d);
	}



	int wd = (int)(152*d);
	int wc = (int)(168*d);
	int w0 = (int)(184*d);
	int w1 = (int) (200*d);
	int w2 = (int)(216*d);
	int w3 = (int)(232*d);
	int w4 = (int)(248*d);
	int w5 = (int)(264*d);
	int w6 = (int)(280*d);
	int w7 = (int)(296*d);
	int w8 = (int)(312*d);
	int w9 = (int)(328*d);
	int w10 = (int)(344*d);
	int w11 = (int)(360*d);

	{
	double m1 = GetEmptyMass();
	double m2 = GetPropellantMass(rcspr);
	double test = ((m1+m2)/m1);
	double dv = (3000*(log(test)));
	skp->SetTextColor(0x0066FF66);
	char kbuf[256];
	sprintf(kbuf,"Remaining dV = %0.2f m/s", dv);
	skp->Text ((10+roxl), (wc+royl), kbuf,strlen(kbuf));
	}

	if (status == 0)
	{
	char abuf[256];
	sprintf(abuf,"J = Jettison SRC Capsule");
	skp->Text ((10+roxl), (w0+royl), abuf,strlen(abuf));
	}
	else
	{
	char bbuf[256];
	sprintf(bbuf,"SRC Capsule Jettisoned");
	skp->Text ((10+roxl), (w0+royl), bbuf,strlen(bbuf));
	}

	if (fold_status == 0)
	{
	char abuf[256];
	sprintf(abuf,"G = Deploy Solar Panels");
	skp->Text ((10+roxl), (w1+royl), abuf,strlen(abuf));
	}
	else
	{
	char bbuf[256];
	sprintf(bbuf,"Solar Panels Deployed");
	skp->Text ((10+roxl), (w1+royl), bbuf,strlen(bbuf));
	}

	if (hinge_status == 0 && status == 0)
	{
	char abuf[256];
	sprintf(abuf,"K = Open SRC Covers");
	skp->Text ((10+roxl), (w2+royl), abuf,strlen(abuf));
	}
	else if (status == 0)
	{
	char bbuf[256];
	sprintf(bbuf,"SRC Covers Opened");
	skp->Text ((10+roxl), (w2+royl), bbuf,strlen(bbuf));
	}

	return true;
}


//void genesis_bus::clbkPostStep (double simt, double simdt, double mjd)
//{
//	
//
//	
//}