// ==============================================================
//                 ORBITER MODULE: genesis_src
//                  Part of the ORBITER SDK
//          Copyright (C) 2002-2004 Martin Schweiger
//                   All rights reserved
//
// genesis_src.cpp
// Control module for genesis_src vessel class
//
// Notes:
// This is an example for a "minimal" vessel implementation which
// only overloads the clbkSetClassCaps method to define vessel
// capabilities and otherwise uses the default VESSEL class
// behaviour.
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"

// ==============================================================
// Some vessel parameters
// ==============================================================


const double CHUTE_OPERATING_SPEED = 1;
const double COL_SPEED = 1;
const double DRG_MASS = 14;
const double MAIN_MASS = 191;
double crtime;
int crashsw = 0;
double acw = 0.9;
double bcw = 0.8;

static const DWORD ntdvtx_geardown = 7;
static TOUCHDOWNVTX tdvtx_geardown[ntdvtx_geardown] = {
	{_V(   0, -0.50,  2), 2e3, 1e4, 2, 2},
	{_V(-1.732,-0.50,-1), 2e3, 1e4, 2, 2},
	{_V( 1.732,-0.50,-1), 2e3, 1e4, 2, 2},
	{_V(   0,  0.5,   0), 1e5, 1e6, 2},
	{_V(    0, -0.50,-2), 2e3, 1e4, 2},
	{_V(-1.732,-0.50, 1), 2e3, 1e4, 2},
	{_V( 1.732,-0.50, 1), 2e3, 1e4, 2},
};

// Calculate lift coefficient [Cl] as a function of aoa (angle of attack) over -Pi ... Pi
// Implemented here as a piecewise linear function
double LiftCoeff (double aoa)
{
	const int nlift = 9;
	static const double AOA[nlift] = {-180*RAD,-60*RAD,-30*RAD,-1*RAD,15*RAD,20*RAD,25*RAD,60*RAD,180*RAD};
	static const double CL[nlift]  = {       0,      0,		0,		0,		0,		0,		0,     0,      0};
	static const double SCL[nlift] = {(CL[1]-CL[0])/(AOA[1]-AOA[0]), (CL[2]-CL[1])/(AOA[2]-AOA[1]),
		                              (CL[3]-CL[2])/(AOA[3]-AOA[2]), (CL[4]-CL[3])/(AOA[4]-AOA[3]),
									  (CL[5]-CL[4])/(AOA[5]-AOA[4]), (CL[6]-CL[5])/(AOA[6]-AOA[5]),
									  (CL[7]-CL[6])/(AOA[7]-AOA[6]), (CL[8]-CL[7])/(AOA[8]-AOA[7])};
	int i;
	for (i = 0; i < nlift-1 && AOA[i+1] < aoa; i++);

	return CL[i] + (aoa-AOA[i])*SCL[i];
}


class genesis_src: public VESSEL3 {
public:
	genesis_src (OBJHANDLE hVessel, int flightmodel);
	~genesis_src ();
	
	
	void clbkSetClassCaps (FILEHANDLE cfg);
	void clbkPreStep(double SimT, double SimDT, double MJD);
	void clbkPostCreation();
	void clbkSaveState (FILEHANDLE scn);
	void clbkLoadStateEx (FILEHANDLE scn, void *status);

	private:

	double CalcEmptyMass();
	void ApplySkin();
	void DoCrash (void);
	void Jettison();
	void Jettison2();
	void SpawnObject2(char* classname2, char* ext2, VECTOR3 ofs2, VECTOR3 vofs2);
	void SpawnAS350();
	void DefineAnimations (void);
	int status, chute_status, col_status;
	int as350_status;
	int blt_sw;
	int ad_sw;
	double prob;
	SURFHANDLE bs1, bs2;//, rtx;
	MESHHANDLE hgen_src_main;
	NOTEHANDLE hNote;
	PROPELLANT_HANDLE fuel;
	THRUSTER_HANDLE th_rcs;
	


	UINT anim_chute, anim_col;
	double chute_proc, col_proc;
	//int status;
};

genesis_src::genesis_src (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
    

}

genesis_src::~genesis_src ()
{
	//if(bs1)delete bs1;
	//delete rtx;
}


// Define animation sequences for moving parts
// --------------------------------------------------------------
void genesis_src::DefineAnimations (void)
{

	// 1. deploy drg chute
	anim_chute = CreateAnimation (0);
	static UINT ChuteGrp[1] = {2};

	static MGROUP_SCALE  ChuteFold1 (0, ChuteGrp, 1, _V(0,0,-0.573527), _V(0.01,0.01,0.001));
	AddAnimationComponent (anim_chute, 0, 1,   &ChuteFold1);
	

	// 2. collapse drg chute
	anim_col = CreateAnimation (0);
	static UINT ColGrp[1] = {2};


	static MGROUP_SCALE  ColFold2 (0, ColGrp, 1, _V(0,0,0), _V(0.8,0.1,1));
	AddAnimationComponent (anim_col, 0.5, 1,   &ColFold2);
	

}


//void genesis_src::RevertChute (void)
//{
//	chute_status = ((chute_status == DOOR_CLOSED || chute_status == DOOR_CLOSING) ?
//		DOOR_OPENING : DOOR_CLOSING);
//}
//
//void genesis_src::RevertCol (void)
//{
//	col_status = ((col_status == DOOR_CLOSED || col_status == DOOR_CLOSING) ?
//		DOOR_OPENING : DOOR_CLOSING);
//}

// ==============================================================
// genesis_src class interface
// ==============================================================

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void genesis_src::clbkSetClassCaps (FILEHANDLE cfg)
{

	chute_proc = 1.0;
	chute_status = 2;
	col_proc = 0.0;
	col_status = 0;
	DefineAnimations ();
	status = 0;
	//as350_status = 0;
	blt_sw = 0;
	ad_sw = 0;
	prob = 0;
	// physical specs
	SetCW (0.9, 0.8, 0.8, 0.8);
	SetSize (3);
	SetCOG_elev (0.1);
	SetCrossSections (_V(2,2,3));
	SetPMI (_V(0.13,0.13,0.20));
	//SetTouchdownPoints (_V(0,-0.3, 1), _V(-1, -0.3, -1), _V(1, -0.3, -1));
	SetTouchdownPoints (tdvtx_geardown,ntdvtx_geardown);
	SetRotDrag (_V(2,2,2));
	if (GetFlightModel() >= 1) {
		SetPitchMomentScale (1e-3);
		SetBankMomentScale (1e-3);
	}
	SetMeshVisibilityMode ((AddMesh ("GenesisMission/gen_src_drg")), MESHVIS_ALWAYS);
	hgen_src_main = oapiLoadMesh("GenesisMission/gen_src_main");
	SetMeshVisibilityMode ((AddMesh (hgen_src_main)), MESHVIS_ALWAYS);
	SetMeshVisibilityMode ((AddMesh ("GenesisMission/gen_src_lid")), MESHVIS_ALWAYS);
	bs2 = oapiLoadTexture ("GenesisMission/gen_src2.dds", true);

	SetCameraOffset (_V(0,1,-2.5));
	SetLiftCoeffFunc (LiftCoeff);

	// reentry pstream
	PARTICLESTREAMSPEC rps = {
		0, 2, 10, 0, 0.001, 0.5, 4, 0.2, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 1, PARTICLESTREAMSPEC::ATM_PLOG, 7e7, 2e8
	};
	AddReentryStream (&rps);
	

	SURFHANDLE rtx = oapiRegisterReentryTexture("contrail3");

	// reentry pstream
	PARTICLESTREAMSPEC rps2 = {
		0, 1, 8, 0, 0.001, 0.5, 4, 0.2, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 1, PARTICLESTREAMSPEC::ATM_PLOG, 9e7, 2e8, rtx
	};
	AddReentryStream (&rps2);


	
	 
	// ***************** thruster definitions *******************
}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new genesis_src (hvessel, flightmodel);
}
// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (genesis_src*)vessel;
}
// ==============================================================
// Overloaded clbkPostCreation callback functions
// ==============================================================

void genesis_src::clbkPostCreation ()
{
	if (status == 0)
	{
	chute_proc = 1.000;
	chute_status = 2;
	SetAnimation (anim_chute, chute_proc);
	}

	if (status > 0 && status < 3) DelMesh(2);
	if (status > 1 && status < 3) DelMesh(1);

	if (status == 4)
	{
		DelMesh(0);
		DelMesh(1);
		DelMesh(2);
		SetMeshVisibilityMode ((AddMesh ("GenesisMission/gen_src_crash")), MESHVIS_ALWAYS);
	}

	prob = GetDamageModel() * oapiRand();

}
void genesis_src::clbkSaveState (FILEHANDLE scn)
{
	char cbuf[256];

	// default vessel parameters
	VESSEL3::clbkSaveState (scn);

	// custom parameters
	sprintf (cbuf, "%d %d", status, as350_status);
	oapiWriteScenario_string (scn, "CONF", cbuf);

	sprintf (cbuf, "%d %0.4f", chute_status, chute_proc);
	oapiWriteScenario_string (scn, "CHUT", cbuf);

	sprintf (cbuf, "%d %0.4f", col_status, col_proc);
	oapiWriteScenario_string (scn, "CLPS", cbuf);
}
void genesis_src::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
    char *line;
	while (oapiReadScenario_nextline (scn, line)) 
	{
		if(!strnicmp (line, "CONF", 4)) 
		{
			sscanf (line+4, "%d%d", &status, &as350_status);
		}
		else if (!strnicmp (line, "CHUT", 4)) 
		{
			sscanf (line+4, "%d%lf", &chute_status, &chute_proc);
		}
		else if (!strnicmp (line, "CLPS", 4))
		{
			sscanf (line+4, "%d%lf", &col_status, &col_proc);
		}
		else
		{
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
   }

	SetAnimation (anim_chute, chute_proc);
	SetAnimation (anim_col, col_proc);
}

void genesis_src::clbkPreStep(double SimT, double SimDT, double MJD) {

	//sprintf(oapiDebugString(),"Prob %f",prob);

	SetEmptyMass(CalcEmptyMass());

	if (crashsw == 1)
	{
	if (oapiGetSimTime() > (crtime + 5)) oapiDelAnnotation (hNote);

	
	acw = acw - (SimDT/10);
	if (acw < 0.3) acw = 0.3;
	bcw = bcw - (SimDT/10);
	if (bcw < 0.2) bcw = 0.2;

	if (25000 > GetAltitude()) SetCW(0.3,0.2,0.2,0.2);
	else SetCW(acw,bcw,bcw,bcw);

	}

	// next block of code
	if (34000 > GetAltitude() && prob > 0.90 && crashsw == 0 && status == 0)
	{
		status = 3; // failure!
		DoCrash();
	}

	if (34000 > GetAltitude() && prob > 0.90 && crashsw == 0 && status == 3)
	{
		DoCrash();
	}

	if (33000 > GetAltitude() && status == 0)
	{
	if (chute_status == 2)chute_status = 1;
	Jettison();
	}

	if (33000 > GetAltitude() && ad_sw < 1 && status < 3)
	{
	SetCrossSections (_V(4,4,6));
	SetRotDrag (_V(0.5,0.5,0.5));
	SetPMI (_V(3,3,3));
	//SetClipRadius (3);
	ad_sw = 1;
	}

	// create as350s + heliports
	if (100000 > GetAltitude() && as350_status < 4) SpawnAS350();

	if (6700 > GetAltitude(ALTMODE_GROUND) && status == 1) Jettison2();
	
	if (6700 > GetAltitude(ALTMODE_GROUND) && ad_sw == 1 && status < 3)
	{
	SetCW (2, 2, 1, 1);
	ad_sw = 2;
	}

	if (1 > GetAltitude(ALTMODE_GROUND) && status < 3)
	{
	if (col_status == 0)col_status = 1;
	}


	if (60000 > GetAltitude() && blt_sw == 0 && status < 2)
	{
		ApplySkin();
		blt_sw = 1;
	}

	if (60000 > GetAltitude() && blt_sw == 0 && status == 3)
	{
		ApplySkin();
		blt_sw = 1;
	}

	DWORD f_st = 1;
	if (GetFlightStatus() == f_st && status == 3)
	{
		DelMesh(0);
		DelMesh(1);
		DelMesh(2);
		SetMeshVisibilityMode ((AddMesh ("GenesisMission/gen_src_crash")), MESHVIS_ALWAYS);
		status = 4;

		
		
		PARTICLESTREAMSPEC crashps = {0, 1.5, 400, 10, 1, 1, -0.2, 0, PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 1, PARTICLESTREAMSPEC::ATM_PLOG, 0, 1};

		fuel = CreatePropellantResource (0.1, 0.1);
		th_rcs = CreateThruster (_V( 0,-0.3,0), _V(0, 1,0), 0.1, fuel, 0.1);
		AddExhaustStream(th_rcs, &crashps);
		SetThrusterLevel(th_rcs,1.0);

	}

	/*if (status == 3)
	{
	VESSELSTATUS2 vs;
	memset (&vs, 0, sizeof(vs));
	vs.version = 2;
	vs.rbody = GetSurfaceRef();
	vs.status = 1;
	OBJHANDLE hVessel = GetHandle();
	double longitude, latitude, radius;
	double heading;
	oapiGetEquPos(hVessel,&longitude,&latitude,&radius);
	oapiGetHeading(hVessel,&heading);
	vs.surf_lat = latitude;
	vs.surf_lng = longitude;
	vs.surf_hdg = heading;
	if (GetAltitude() < 2) DefSetStateEx (&vs);
	}*/

		// Animate chutes
	if (chute_status == 0) chute_proc = 0;
	else if (chute_status == 1)
	{
		double da = SimDT * CHUTE_OPERATING_SPEED;
		chute_proc = chute_proc - da;
		if (chute_proc < 0)
		{
			chute_proc = 0;
			chute_status = 0;
		}
	}
	else if (chute_status == 2) chute_proc = 1;
	SetAnimation (anim_chute, chute_proc);
	

	// collapse chutes
	if (col_status == 0) col_proc = 0;
	else if (col_status == 1)
	{
		double da = SimDT * CHUTE_OPERATING_SPEED;
		col_proc = col_proc + da;
		if (col_proc > 1)
		{
			col_proc = 1;
			col_status = 2;
		}
	}
	else if (col_status == 2) col_proc = 1;
	SetAnimation (anim_col, col_proc);
	
}


double genesis_src::CalcEmptyMass() 
{
	double EmptyMass=0;
	if (status < 2) {
	EmptyMass+=MAIN_MASS;
	}

	if (status == 3) {
	EmptyMass+=MAIN_MASS;
	}

	EmptyMass+=DRG_MASS;
	return EmptyMass;
} 

void genesis_src::ApplySkin ()
{	
	 bs1 = oapiGetTextureHandle (hgen_src_main, 4);
	oapiBlt(bs1, bs2, 0, 0, 0, 0, 512, 512);
}

void genesis_src::DoCrash (void)
{

if (crashsw == 0)
	{
crtime = oapiGetSimTime();
crashsw = 1;
SetRotDrag (_V(0,0,0));
if (GetFlightModel() >= 1) {
		SetPitchMomentScale (0);
		SetBankMomentScale (0);
	}
VECTOR3 col = _V(1,0,0);
hNote = oapiCreateAnnotation(true, 2, col);
oapiAnnotationSetPos(hNote,0.3,0.4,0.8,0.8);
oapiAnnotationSetText(hNote, "Chute Failure!");
	}

}

void genesis_src::Jettison() {
  switch(status) {
    case 0:
      status=1;
	  DelMesh(2);
      break;
  }
}


void genesis_src::Jettison2() {
  switch(status) {
    case 1:
      status=2;
	  DelMesh(1);
      SpawnObject2("GenesisMission/genesis_src_parafoil","Genesis_SRC_Parafoil",_V(0,0,0),_V(0,0,0));
	  //oapiDeleteVessel(GetHandle());
      break;
  }
}

void genesis_src::SpawnObject2(char* classname2, char* ext2, VECTOR3 ofs2, VECTOR3 vofs2) {

	VECTOR3 pos;
	VECTOR3 vel;
	GlobalRot (_V(1,0,0), pos);
	GlobalRot (_V(0,1,0), vel);
	VECTOR3 yup;
	VECTOR3 yupn;
	VECTOR3 zforward;
	VECTOR3 xright;
	yupn = crossp(vel, pos);
	double yupnlength;
	yupnlength = length(yupn);
	yup = _V(yupn.x/yupnlength, yupn.y/yupnlength, yupn.z/yupnlength);
	xright = crossp (yup, vel);
	zforward = vel;
	
	double gamma = atan2( xright.x, yup.x);
	double beta = -1*(asin(zforward.x));
	double alpha = atan2(zforward.y, zforward.z);

  VESSELSTATUS vs;
  char name[256];
  GetStatus(vs);
  VECTOR3 rvel;
  VECTOR3 angvel = vs.vrot;
  Local2Rel (ofs2, vs.rpos);
  GlobalRot (vofs2, rvel);
  vs.rvel = vs.rvel + rvel;
  vs.arot = _V(alpha, beta, (PI/2)-gamma);
  vs.vrot.x = vs.vrot.x;
  vs.vrot.y = -angvel.z;
  vs.vrot.z = angvel.y;
  vs.eng_main = vs.eng_hovr = 0.0;
  vs.status = 0;
  strcpy (name, ext2);
  OBJHANDLE hVes = oapiCreateVessel (name, classname2, vs);
  //oapiSetFocusObject(hVes);
}


void genesis_src::SpawnAS350()
{
	double longitude, latitude, alti;
	VECTOR3 arot_vect;
	char cfg_path[256];
	char name[256];

	if (as350_status == 0)
	{
	longitude = -113.7842200*RAD;
	latitude = 40.4497800*RAD;
	alti = 0.010;
	arot_vect = _V(52.040*RAD, -15.168*RAD, -18.540*RAD);
	strcpy(cfg_path,"GenesisMission/heliport");
	strcpy(name,"Heliport_1");
	as350_status = 1;
	}
	else if (as350_status == 1)
	{
	longitude = -113.7842200*RAD;
	latitude = 40.4497800*RAD;
	alti = 1.443;
	arot_vect = _V(52.022*RAD, -15.174*RAD, -18.536*RAD);
	strcpy(cfg_path,"GenesisMission/as350");
	strcpy(name,"AS350_1");
	as350_status = 2;
	}
	else if (as350_status == 2)
	{
	longitude = -113.2833000*RAD;
	latitude = 39.9500000*RAD;
	alti = 0.010;
	arot_vect = _V(-127.575*RAD, 14.703*RAD, -161.743*RAD);
	strcpy(cfg_path,"GenesisMission/heliport");
	strcpy(name,"Heliport_2");
	as350_status = 3;
	}
	else if (as350_status == 3)
	{
	longitude = -113.2833000*RAD;
	latitude = 39.9500000*RAD;
	alti = 1.443;
	arot_vect = _V(-127.579*RAD, 14.631*RAD, -161.744*RAD);
	strcpy(cfg_path,"GenesisMission/as350");
	strcpy(name,"AS350_2");
	as350_status = 4;
	}

	VESSELSTATUS2 vs;
	memset (&vs, 0, sizeof(vs));
	vs.version = 2;
	vs.rbody = GetSurfaceRef(); // the Earth, usually!
	vs.status = 1; // status landed
	vs.surf_lat = latitude;
	vs.surf_lng = longitude;
	vs.vrot.x = alti;
	vs.vrot.y = 0;
	vs.vrot.z = 0;
	vs.arot = arot_vect;
//	vs.surf_hdg = PI;

	OBJHANDLE hVes = oapiCreateVesselEx(name, cfg_path, &vs);

}

