// ==============================================================
//                 ORBITER MODULE: genesis_src_parafoil
//                  Part of the ORBITER SDK
//          Copyright (C) 2002-2004 Martin Schweiger
//                   All rights reserved
//
// genesis_src_parafoil.cpp
// Control module for genesis_src_parafoil vessel class
//
// Notes:
// This is an example for a "minimal" vessel implementation which
// only overloads the clbkSetClassCaps method to define vessel
// capabilities and otherwise uses the default VESSEL class
// behaviour.
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"

// ==============================================================
// Some vessel parameters
// ==============================================================


const double COL_SPEED = 0.5;

static const DWORD ntdvtx_geardown = 7;
static TOUCHDOWNVTX tdvtx_geardown[ntdvtx_geardown] = {
	{_V(   0, -0.50,  2), 2e3, 1e4, 2, 2},
	{_V(-1.732,-0.50,-1), 2e3, 1e4, 2, 2},
	{_V( 1.732,-0.50,-1), 2e3, 1e4, 2, 2},
	{_V(   0,  0.5,   0), 1e5, 1e6, 2},
	{_V(    0, -0.50,-2), 2e3, 1e4, 2},
	{_V(-1.732,-0.50, 1), 2e3, 1e4, 2},
	{_V( 1.732,-0.50, 1), 2e3, 1e4, 2},
};

// Calculate lift coefficient [Cl] as a function of aoa (angle of attack) over -Pi ... Pi
// Implemented here as a piecewise linear function
void VLiftCoeff (VESSEL *v, double aoa, double M, double Re, void *context, double *cl, double *cm, double *cd)
{
	int i;
	const int nabsc = 9;
	static const double AOA[nabsc] = {-180*RAD,-60*RAD,-30*RAD, -2*RAD, 10*RAD,15*RAD,50*RAD,90*RAD,180*RAD};
	//static const double AOA[nabsc] = {-180*RAD,-60*RAD,-30*RAD, -2*RAD, 10*RAD,15*RAD,20*RAD,60*RAD,180*RAD};
	static const double CL[nabsc]  = {       0,      0,   -0.4,      0,    0.3,     0.6,   0.8,     0,      0};
	//static const double CL[nabsc]  = {       0,      0,			0,      0,    0.7,     1,   0.8,     0,      0};
	static const double CM[nabsc]  = {       0,      0,  0.014, 0.0039, -0.006,-0.008,-0.010,     0,      0};
	for (i = 0; i < nabsc-1 && AOA[i+1] < aoa; i++);
	double f = (aoa-AOA[i]) / (AOA[i+1]-AOA[i]);
	*cl = CL[i] + (CL[i+1]-CL[i]) * f;  // aoa-dependent lift coefficient
	*cm = CM[i] + (CM[i+1]-CM[i]) * f;  // aoa-dependent moment coefficient
	double saoa = sin(aoa);
	double pd = 0.03 + 0.4*saoa*saoa;  // profile drag
	*cd = pd + oapiGetInducedDrag (*cl, 1.5, 0.7) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
	// profile drag + (lift-)induced drag + transonic/supersonic wave (compressibility) drag
}

// 2. horizontal lift component (vertical stabilisers and body)

void HLiftCoeff (VESSEL *v, double beta, double M, double Re, void *context, double *cl, double *cm, double *cd)
{
	int i;
	const int nabsc = 8;
	static const double BETA[nabsc] = {-180*RAD,-135*RAD,-90*RAD,-45*RAD,45*RAD,90*RAD,135*RAD,180*RAD};
	static const double CL[nabsc]   = {       0,    +0.3,      0,   -0.4,  +0.4,     0,   -0.3,      0};
	for (i = 0; i < nabsc-1 && BETA[i+1] < beta; i++);
	*cl = CL[i] + (CL[i+1]-CL[i]) * (beta-BETA[i]) / (BETA[i+1]-BETA[i]);
	*cm = 0.0;
	*cd = 0.03 + oapiGetInducedDrag (*cl, 1.5, 0.6) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
}



class genesis_src_parafoil: public VESSEL2 {
public:
	enum DoorStatus { DOOR_CLOSED, DOOR_OPEN, DOOR_CLOSING, DOOR_OPENING } col_status;
	genesis_src_parafoil (OBJHANDLE hVessel, int flightmodel)
		: VESSEL2 (hVessel, flightmodel) {}
	void DefineAnimations (void);
	void RevertCol (void);
	void DoAttAngle (void);
	void clbkSetClassCaps (FILEHANDLE cfg);
	void clbkPreStep(double SimT, double SimDT, double MJD);
	void clbkPostStep (double simt, double simdt, double mjd);
	void clbkPostCreation();
	void clbkSaveState (FILEHANDLE scn);
	void clbkLoadStateEx (FILEHANDLE scn, void *status);
	MESHHANDLE hgen_src_para_col1, hgen_src_para_col2;
	AIRFOILHANDLE hwing;
	ATTACHMENTHANDLE hatp;
	int ad_sw;
	int status;
	double  col_proc;

private:
	UINT  anim_col;
	//double  col_proc;

};
// Define animation sequences for moving parts
// --------------------------------------------------------------
void genesis_src_parafoil::DefineAnimations (void)
{

	
	// 2. collapse drg chute
	anim_col = CreateAnimation (0);
	static UINT ColGrp[1] = {5};

	static MGROUP_SCALE  ColFold2 (0, ColGrp, 1, _V(0,0,0), _V(0.25,0.04,0.25));
	AddAnimationComponent (anim_col, 0.1, 1,   &ColFold2);

	static MGROUP_SCALE  ColFold3 (0, ColGrp, 1, _V(0,0,0), _V(0.2,1,0.2));
	AddAnimationComponent (anim_col, 0, 0.3,   &ColFold3);
	

}



void genesis_src_parafoil::RevertCol (void)
{
	col_status = ((col_status == DOOR_CLOSED || col_status == DOOR_CLOSING) ?
		DOOR_OPENING : DOOR_CLOSING);
}

// ==============================================================
// genesis_src_parafoil class interface
// ==============================================================

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void genesis_src_parafoil::clbkSetClassCaps (FILEHANDLE cfg)
{

	col_proc = 1.0;
	col_status = DOOR_OPEN;
	DefineAnimations ();
	ad_sw = 0;
	status = 0;
	// physical specs
	SetCW (2, 0.8, 0.8, 0.8);
	SetSize (7);
	SetCrossSections (_V(9,52,35));
	SetPMI (_V(88,7,88));
	SetEmptyMass(191);
	SetTouchdownPoints (tdvtx_geardown,ntdvtx_geardown);
	//SetRotDrag (_V(2,2,2));
	if (GetFlightModel() >= 1) {
		SetPitchMomentScale (1e-3);
		SetBankMomentScale (1e-3);
	}
	SetMeshVisibilityMode ((AddMesh ("GenesisMission/gen_src_parafoil90")), MESHVIS_ALWAYS);
	hgen_src_para_col1 = oapiLoadMesh("GenesisMission/gen_src_parafoil_col1");
	hgen_src_para_col2 = oapiLoadMesh("GenesisMission/gen_src_parafoil_col2");


	SetCameraOffset (_V(0,1,-4));
	
	hatp = CreateAttachment (true, _V(0,13,0), _V(0,0,-1), _V(0,1,0), "GEN");
	// ********************* aerodynamics ***********************

	hwing = CreateAirfoil3 (LIFT_VERTICAL, _V(0.10,3.5,-0.3), VLiftCoeff, 0, 5, 90, 1.5);
	// wing and body lift+drag components

	CreateAirfoil3 (LIFT_HORIZONTAL, _V(0,1,-4), HLiftCoeff, 0, 5, 15, 1.5);
	// vertical stabiliser and body lift and drag components

	CreateControlSurface (AIRCTRL_ELEVATORTRIM, 0.3, 1.5, _V(   0,0,-7.2), AIRCTRL_AXIS_XPOS);

	SetRotDrag (_V(0.7,0.7,0.7));
	// angular damping

}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new genesis_src_parafoil (hvessel, flightmodel);
}
// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (genesis_src_parafoil*)vessel;
}
// ==============================================================
// Overloaded clbkPostCreation callback functions
// ==============================================================

void genesis_src_parafoil::clbkPostCreation ()
{
		if (status == 1)
		{
		DelMesh(0);
		SetMeshVisibilityMode ((AddMesh (hgen_src_para_col1)), MESHVIS_ALWAYS);
		SetCrossSections (_V(2,2,2));
		SetCW (0.5, 0.5, 0.5, 0.5);
		ClearAirfoilDefinitions ();
		ClearControlSurfaceDefinitions ();
		}

		if (status == 2)
		{
		DelMesh(0);
		DelMesh(1);
		SetMeshVisibilityMode ((AddMesh (hgen_src_para_col2)), MESHVIS_ALWAYS);
		ClearAirfoilDefinitions ();
		ClearControlSurfaceDefinitions ();
		SetCrossSections (_V(2,2,2));
		SetCW (0.2, 0.2, 0.2, 0.2);
		SetPMI (_V(2,2,2));
		}

}
void genesis_src_parafoil::clbkSaveState (FILEHANDLE scn)
{
	char cbuf[256];

	// default vessel parameters
	VESSEL2::clbkSaveState (scn);

	// custom parameters
	oapiWriteScenario_int (scn, "CONFIGURATION", status);
	sprintf (cbuf, "%d %0.4f", col_status, col_proc);
	oapiWriteScenario_string (scn, "COL", cbuf);
}
void genesis_src_parafoil::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
    char *line;
	while (oapiReadScenario_nextline (scn, line)) {
		if(!strnicmp (line, "CONFIGURATION", 13)) {
            sscanf (line+13, "%d", &status);
		} else if (!strnicmp (line, "COL", 3)) {
			sscanf (line+3, "%d%lf", &col_status, &col_proc);
		}else {
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
   }

	SetAnimation (anim_col, col_proc);
}

void genesis_src_parafoil::clbkPreStep(double SimT, double SimDT, double MJD) {

	if (col_status == DOOR_OPEN) RevertCol();
	
	SetControlSurfaceLevel(AIRCTRL_ELEVATORTRIM, 1.0);

	if (GetAttachmentStatus(hatp) == NULL);
	else if (status == 0)
	{
	DelMesh(0);
	DelMesh(1);
	SetMeshVisibilityMode ((AddMesh (hgen_src_para_col1)), MESHVIS_ALWAYS);
	status = 1;
	SetCrossSections (_V(2,2,2));
	SetCW (0.5, 0.5, 0.5, 0.5);
	ClearAirfoilDefinitions ();
	ClearControlSurfaceDefinitions ();
	}

	/*if (GetAttachmentStatus(hatp) == NULL && status == 1)
	{
	DelMesh(0);
	DelMesh(1);
	SetMeshVisibilityMode ((AddMesh ("GenesisMission/gen_src_parafoil90")), MESHVIS_ALWAYS);
	status = 0;
	}*/

	if (status == 1) DoAttAngle();

	//DWORD f_st = 1;
	//if (GetFlightStatus() == f_st && status < 2)
	if(GroundContact() == true)
	{
		DelMesh(0);
		DelMesh(1);
		SetMeshVisibilityMode ((AddMesh (hgen_src_para_col2)), MESHVIS_ALWAYS);
		status = 2;
		ClearAirfoilDefinitions ();
		ClearControlSurfaceDefinitions ();
		SetCrossSections (_V(2,2,2));
		SetPMI (_V(2,2,2));
		SetCW (0.2, 0.2, 0.2, 0.2);
	}

// lock down to ground
//if (GetAltitude() < 0.5)
//{
//	VESSELSTATUS2 vs;
//	memset (&vs, 0, sizeof(vs));
//	vs.version = 2;
//	vs.rbody = GetSurfaceRef();
//	vs.status = 1;
//	OBJHANDLE hVessel = GetHandle();
//	double longitude, latitude, radius;
//	double heading;
//	oapiGetEquPos(hVessel,&longitude,&latitude,&radius);
//	oapiGetHeading(hVessel,&heading);
//	vs.surf_lat = latitude;
//	vs.surf_lng = longitude;
//	vs.surf_hdg = heading;
//	DefSetStateEx (&vs);
//}
	
}

void genesis_src_parafoil::DoAttAngle (void)
{
VECTOR3 aspdvec;
GetAirspeedVector(FRAME_LOCAL, aspdvec);
double aspd = aspdvec.z;

if (aspd > 450) aspd = 450;
if (aspd < -450) aspd = -450;
aspd = aspd/5;

double angy;
double angz;
angy = -sin(aspd*RAD);
angz = -cos(aspd*RAD);

SetAttachmentParams(hatp,_V(0,11,0),_V(0,angy,angz),_V(0,-angz,angy));


}
void genesis_src_parafoil::clbkPostStep (double simt, double simdt, double mjd)
{

	// collapse chutes
	if (col_status >= DOOR_CLOSING) {
		double da = simdt * COL_SPEED;
		if (col_status == DOOR_CLOSING) {
			if (col_proc > 0.0) col_proc = max (0.0, col_proc-da);
			else                col_status = DOOR_CLOSED;
		} else {
			if (col_proc < 1.0) col_proc = min (1.0, col_proc+da);
			else                col_status = DOOR_OPEN;
		}
		SetAnimation (anim_col, col_proc);
	}
}